#!/usr/bin/env python3.8
import argparse
import logging
import pathlib
import re
import typing
from dataclasses import dataclass, field
from typing import Dict, Generator, List, Optional, Tuple, Union

import apischema
import lark

from . import transformer
from .common import FullLoadContext, StringWithContext, context_from_lark_token
from .util import get_bytes_sha256, get_file_sha256

MODULE_PATH = pathlib.Path(__file__).parent.resolve()
logger = logging.getLogger(__name__)


@dataclass
class Rule:
    """A PVList rule (base class)."""
    context: FullLoadContext
    pattern: str
    command: str
    regex: typing.Pattern = field(
        default=None,
        metadata=apischema.metadata.skip
    )
    header: str = ""
    metadata: Dict[str, str] = field(default_factory=dict)

    def __post_init__(self):
        if self.regex is None:
            try:
                self.regex = re.compile(self.pattern)
            except Exception as ex:
                self.metadata["error"] = (
                    f"Invalid regex. {ex.__class__.__name__}: {ex}"
                )

    def match(self, name):
        """Match a pv name against this rule."""
        if self.regex is not None:
            return self.regex.fullmatch(name)


@dataclass
class AccessSecurity:
    """A PVList rule access security settings."""
    group: Optional[str] = None
    level: Optional[str] = None


@dataclass
class AliasRule(Rule):
    """Rule to alias the pattern to a PV (or PVs)."""
    pvname: str = ""
    access: Optional[AccessSecurity] = None


@dataclass
class DenyRule(Rule):
    """Rule to deny access to a PV pattern."""
    hosts: List[str] = field(default_factory=list)


@dataclass
class AllowRule(Rule):
    """Rule to allow access to a PV pattern."""
    access: Optional[AccessSecurity] = None


@lark.visitors.v_args(inline=True)
class _PVListTransformer(lark.visitors.Transformer_InPlaceRecursive):
    fn: str
    contents_hash: str
    comments: List[str]

    def __init__(self, fn, contents_hash: str, comments: List[str]):
        super().__init__(visit_tokens=False)
        self.fn = str(fn)
        self.contents_hash = contents_hash
        self.comments = comments

    def get_header_comments(self) -> str:
        """Get all comments at the top of the file."""
        results = []
        lineno = 0
        for comment in self.comments:
            if comment.context[0].line == lineno + 1:
                results.append(str(comment).lstrip("# "))
                lineno += 1
            else:
                break

        return "\n".join(results)

    def get_recent_comments(self, lineno: int) -> str:
        """Get all comments just before the given line."""
        results = []
        for comment in reversed(self.comments):
            comment_line = comment.context[0].line
            if comment_line == lineno - 1:
                results.append(str(comment).lstrip("# "))
                lineno -= 1
            elif comment_line > lineno:
                # lexer can get ahead of this line; keep going.
                ...
            else:
                break

        return "\n".join(reversed(results))

    def pvlist(self, *items):
        rules = []
        evaluation = None
        for item in items:
            if isinstance(item, Rule):
                rules.append(item)
            elif isinstance(item, str):
                evaluation = item
            else:
                raise ValueError("unexpected top-level item?")

        return PVList(
            filename=self.fn,
            evaluation_order=evaluation,
            rules=list(rules),
            header=self.get_header_comments(),
            hash=self.contents_hash,
        )

    def evaluation_order(self, first, _, second) -> str:
        return f"{first.upper()}, {second.upper()}"

    def evaluation(self, _evaluation, _order, order, *_):
        return order

    pattern = transformer.stringify
    pvname = transformer.stringify
    hosts = transformer.listify
    host = transformer.stringify

    def allow(
        self,
        pattern: str,
        allow_token: lark.Token,
        asg_asl: Optional[AccessSecurity] = None,
        *_
    ) -> AllowRule:
        return AllowRule(
            context=context_from_lark_token(self.fn, allow_token),
            command=str(allow_token).upper(),
            pattern=pattern,
            access=asg_asl if _ else None,
            header=self.get_recent_comments(allow_token.line),
        )

    def alias(
        self,
        pattern: str,
        alias_token: lark.Token,
        pvname: str,
        asg_asl: Optional[AccessSecurity] = None,
        *_
    ) -> AliasRule:
        return AliasRule(
            context=context_from_lark_token(self.fn, alias_token),
            command=str(alias_token).upper(),
            pattern=pattern,
            pvname=pvname,
            access=asg_asl if _ else None,
            header=self.get_recent_comments(alias_token.line),
        )

    def deny(
        self,
        pattern: str,
        deny_token: lark.Token,
        from_hosts: Optional[List[str]] = None,
        *_
    ) -> DenyRule:
        hosts = None
        if from_hosts and str(from_hosts).strip():
            hosts = from_hosts

        return DenyRule(
            context=context_from_lark_token(self.fn, deny_token),
            command=str(deny_token).upper(),
            pattern=pattern,
            hosts=hosts or [],
            header=self.get_recent_comments(deny_token.line),
        )

    def asg_asl(self, group: str, level: Optional[str] = None) -> AccessSecurity:
        return AccessSecurity(group=group, level=level)

    asg = transformer.stringify
    asl = transformer.stringify


@dataclass
class PVList:
    """A PVList container."""
    filename: Optional[str] = None
    evaluation_order: Optional[str] = None
    rules: List[Rule] = field(default_factory=list)
    hash: Optional[str] = None
    header: str = ""
    comments: List[lark.Token] = field(
        default_factory=list,
        metadata=apischema.metadata.skip,
    )

    def find(
        self, cls: typing.Type
    ) -> Generator[Rule, None, None]:
        """Yield matching rule types."""
        for rule in self.rules:
            if isinstance(rule, cls):
                yield rule

    def match(
        self, name: str
    ) -> Generator[Tuple[Rule, List[str]], None, None]:
        """Yield matching rules."""
        for rule in self.rules:
            m = rule.match(name)
            if m:
                yield rule, list(m.groups())

    @classmethod
    def from_string(cls, contents: str, filename: Optional[str] = None):
        contents_hash = get_bytes_sha256(contents.encode("utf-8"))
        comments = []

        def add_comment(comment: lark.Token):
            comments.append(
                StringWithContext(
                    str(comment).lstrip("# "),
                    context=context_from_lark_token(filename, comment),
                )
            )

        grammar = lark.Lark.open_from_package(
            "whatrecord",
            "gateway.lark",
            search_paths=("grammar", ),
            parser="lalr",
            maybe_placeholders=False,
            lexer_callbacks={"COMMENT": add_comment},
            transformer=_PVListTransformer(filename, contents_hash, comments),
        )

        # Sorry, the grammar isn't perfect: require a newline for rules
        pvlist: PVList = grammar.parse(f"{contents.strip()}\n")
        pvlist.comments = comments
        return pvlist

    @classmethod
    def from_file_obj(cls, fp, filename: Optional[str] = None):
        """Load a PVList from a file object."""
        filename = filename or getattr(fp, "name", str(id(fp)))
        return cls.from_string(fp.read(), filename=filename)

    @classmethod
    def from_file(cls, fn: Union[str, pathlib.Path]):
        """Load a PVList from a filename."""
        with open(fn, "rt") as fp:
            return cls.from_file_obj(fp, filename=str(fn))


def create_arg_parser():
    """Create the argument parser."""
    parser = argparse.ArgumentParser(
        description="pvlist name matching and linting tool",
    )
    parser.add_argument("--lint", action="store_true", help="Lint regular expressions")
    parser.add_argument(
        "--pvlists",
        type=str,
        nargs="*",
        help="Specific pvlists to check (empty for all)",
    )
    parser.add_argument(
        "--hide-context", action="store_true", help="Hide comment context"
    )
    parser.add_argument(
        "--remove-any", action="store_true", help="Remove '.*' from results"
    )
    parser.add_argument("names", nargs="*", type=str, help="PV names to match")
    return parser


@dataclass
class PVListMatch:
    filename: str
    rule: Rule
    groups: List[str]


@dataclass
class PVListMatches:
    name: str
    matches: List[PVListMatch]


@dataclass
class GatewayConfig:
    path: pathlib.Path
    glob_str: str
    pvlists: Dict[pathlib.Path, PVList]

    def __init__(
        self,
        path: Union[str, pathlib.Path],
        glob_str: str = "*.pvlist",
        pvlists: Optional[Dict[pathlib.Path, PVList]] = None,
    ):
        self.path = pathlib.Path(path).resolve()
        self.glob_str = glob_str
        if pvlists is not None:
            self.pvlists = pvlists
        else:
            self.pvlists = self._load_from_files()

    def _load_from_files(self) -> Dict[pathlib.Path, PVList]:
        """Load all PVList files from the current path/glob settings."""
        if self.path.is_file():
            filenames = [self.path]
        else:
            filenames = [p.resolve() for p in self.path.glob(self.glob_str)]
        return {
            filename: PVList.from_file(filename) for filename in filenames
        }

    def _update(self, filename):
        """Update a gateway configuration file."""
        self.pvlists[filename] = PVList.from_file(filename)

    def update_changed(self):
        """Update any changed files."""
        for filename, pvlist in self.pvlists.items():
            if get_file_sha256(filename) != pvlist.hash:
                logger.info("Updating changed gateway file: %s", filename)
                self._update(filename)

    def get_matches(self, name: str, remove_any: bool = True) -> PVListMatches:
        """Get matches from any PVList given a PV name."""
        matches = [
            PVListMatch(
                filename=str(fn),
                rule=rule,
                groups=groups,
            )
            for fn, pvlist in self.pvlists.items()
            for rule, groups in pvlist.match(name)
            if rule.pattern != ".*" or not remove_any
        ]

        return PVListMatches(
            name=name,
            matches=matches,
        )



"""
"whatrecord lint" is used to lint a startup script or database file.
"""

import argparse
import logging
import pathlib
import sys
from typing import IO, List, Optional, Union

from ..common import IocshCommand
from ..db import Database
from ..format import FormatContext
from ..shell import LoadedIoc
from .parse import parse_from_cli_args

logger = logging.getLogger(__name__)
DESCRIPTION = __doc__


def build_arg_parser(parser=None):
    if parser is None:
        parser = argparse.ArgumentParser()

    parser.description = DESCRIPTION
    parser.formatter_class = argparse.RawTextHelpFormatter

    parser.add_argument(
        'filename',
        type=str,
        help="Startup script filename"
    )

    parser.add_argument(
        '--dbd',
        type=str,
        help="The dbd file, if parsing a database",
    )

    parser.add_argument(
        '--standin-directory',
        type=str,
        nargs='*',
        help='Map a "stand-in" directory to another on disk'
    )

    parser.add_argument(
        '--json',
        dest="as_json",
        action="store_true",
    )

    parser.add_argument(
        '-v', '--verbose',
        dest="verbosity",
        action="count",
        help="Increase verbosity"
    )

    parser.add_argument(
        '--use-gdb',
        action="store_true",
        help="Use metadata derived from the script binary",
    )

    return parser


def lint_command(
    command_info: IocshCommand,
    argv: List[str],
    file: IO[str] = sys.stdout,
    fmt: FormatContext = None,
):
    """Lint a command given its argument information."""
    if not command_info:
        print(f"  ! Warning: Unknown command: {argv[0]}")
        return

    expected_args = command_info.args
    actual_args = argv[1:]
    arg_names = [arg.name for arg in expected_args]
    if len(actual_args) == len(expected_args):
        return

    if len(actual_args) < len(expected_args):
        arg_names = [arg.name for arg in expected_args]
        arg_values = list(actual_args) + ["?"] * (len(expected_args) - len(actual_args))
        print("  ! Warning: may be too few arguments", file=file)
    else:
        print("  ! Warning: too many arguments", file=file)
        arg_names = [
            arg.name for arg in expected_args
        ] + ["?"] * (len(actual_args) - len(expected_args))
        arg_values = actual_args

    for idx, (arg_name, value) in enumerate(zip(arg_names, arg_values), 1):
        print(f"    {idx}. {arg_name} = {value}", file=file)
    print(file=file)


def lint(
    obj: Union[Database, LoadedIoc],
    *,
    verbosity: int = 0,
    file: IO[str] = sys.stdout,
    fmt: FormatContext = None,
) -> Union[Database, LoadedIoc]:
    """
    Lint a startup script or a database file.

    Parameters
    ----------
    obj :
        The object to lint.
    """
    fmt = fmt or FormatContext()
    if isinstance(obj, LoadedIoc):
        commands = obj.metadata.commands
        # variables = obj.metadata.variables
        for line in obj.script.lines:
            if line.line or verbosity > 2:
                if line.error or verbosity > 1:
                    print(fmt.render_object(line).rstrip(), file=file)
                if commands and line.argv:
                    command_info = commands.get(line.argv[0], None)
                    lint_command(command_info=command_info, argv=line.argv, file=file, fmt=fmt)


def main(
    filename: Union[str, pathlib.Path],
    dbd: Optional[str] = None,
    standin_directory: Optional[List[str]] = None,
    macros: Optional[str] = None,
    as_json: bool = False,
    use_gdb: bool = False,
    verbosity: Optional[int] = 0,
):
    obj = parse_from_cli_args(
        filename=filename,
        dbd=dbd,
        standin_directory=standin_directory,
        macros=macros,
        use_gdb=use_gdb,
    )
    lint(obj, verbosity=verbosity or 0)



from dataclasses import dataclass
from typing import Dict

from ..cache import Cached, CacheKey, InlineCached


def test_simple(tmp_path):
    @dataclass
    class TestKey(CacheKey):
        a: str
        b: int

    @dataclass
    class Test(Cached, key=TestKey, version=1, cache_path=tmp_path):
        a: str

    key = TestKey(a="a", b=2)
    obj1 = Test(key=key, a="1")
    assert obj1.save_to_cache()

    fn = key.to_filename(version=1, class_name="Test")
    assert (tmp_path / fn).exists()

    with open(tmp_path / fn) as fp:
        print("Cached data:")
        print(fp.read())

    obj2 = Test.from_cache(key)
    assert obj2 is not None
    print(obj1, obj2)
    assert obj1 == obj2


def test_dict_key():
    @dataclass
    class TestKey(CacheKey):
        a: dict
        b: int

    # dict key ordering shouldn't matter:
    key1 = TestKey(a={"a": 1, "b": 2}, b=2)
    key2 = TestKey(a={"b": 2, "a": 1}, b=2)
    assert key1.to_filename() == key2.to_filename()

    key3 = TestKey(a={"b": 2, "a": 1}, b=3)
    assert key3 != key1


def test_similar_keys():
    @dataclass
    class TestKey1(CacheKey):
        a: int
        b: int

    @dataclass
    class TestKey2(CacheKey):
        a: int
        b: int

    key1 = TestKey1(a=1, b=2)
    key2 = TestKey2(a=1, b=2)
    assert key1.to_filename() != key2.to_filename()


def test_identical_keys():
    @dataclass
    class TestKey1(CacheKey):
        a: int
        b: int

    @dataclass
    class TestKey2(CacheKey):
        a: int
        b: int

    key1 = TestKey1(a=1, b=2)
    key2 = TestKey2(a=1, b=2)

    # There's no way around this; so be careful with your CacheKey usage
    assert key1.to_filename(class_name="test") == key2.to_filename(class_name="test")


def test_simple_inline(tmp_path):
    @dataclass
    class TestKey(CacheKey):
        a: str
        b: int

    @dataclass
    class Test(InlineCached, TestKey, version=1, cache_path=tmp_path):
        ...

    key = TestKey(a="a", b=2)
    obj1 = Test(a="a", b=2)

    assert key.to_filename(class_name="Test") == obj1.to_filename()

    assert obj1.save_to_cache()

    fn = key.to_filename(version=1, class_name="Test")
    assert (tmp_path / fn).exists()

    with open(tmp_path / fn) as fp:
        print("Cached data:")
        print(fp.read())

    obj2 = Test.from_cache(key)
    assert obj2 is not None
    print("loaded from cache", obj2)
    assert obj1 == obj2


def test_inline_with_dataclass(tmp_path):
    @dataclass
    class Dependency:
        name: str

    @dataclass
    class TestKey(CacheKey):
        a: Dependency
        b: Dict[str, Dependency]

    @dataclass
    class Test(InlineCached, TestKey, version=1, cache_path=tmp_path):
        ...

    key = TestKey(a=Dependency(name="foo"), b={"c": Dependency(name="bar")})
    obj1 = Test(a=Dependency(name="foo"), b={"c": Dependency(name="bar")})

    assert key.to_filename(class_name="Test") == obj1.to_filename()

    assert obj1.save_to_cache()

    fn = key.to_filename(version=1, class_name="Test")
    assert (tmp_path / fn).exists()

    with open(tmp_path / fn) as fp:
        print("Cached data:")
        print(fp.read())

    obj2 = Test.from_cache(key)
    assert obj2 is not None
    print("loaded from cache", obj2)
    assert obj1 == obj2



"""Tests for the 'whatrecord lint' subcommand."""

import os

import pytest

from ..bin.lint import main
from . import conftest

scripts_with_errors = {"ioc_failure", }


@conftest.startup_scripts
def test_load_no_errors(startup_script, capsys):
    if startup_script.parent.name in scripts_with_errors:
        pytest.skip(f"{startup_script.parent.name} is excluded from this test")

    os.environ["PWD"] = str(startup_script.resolve().parent)
    main(startup_script, verbosity=2)
    captured = capsys.readouterr()
    print("stdout:", captured.out)
    print("stderr:", captured.err)

    assert not captured.err
    assert "ERROR" not in captured.out



from __future__ import annotations

import inspect
import json
import logging
import pathlib
from dataclasses import asdict, dataclass, field, fields, is_dataclass
from typing import Any, ClassVar, Mapping, Optional, Sequence, Type, TypeVar

import apischema

from whatrecord.common import AnyPath
from whatrecord.settings import CACHE_PATH
from whatrecord.util import get_bytes_sha256

T = TypeVar("T")


logger = logging.getLogger(__name__)


@dataclass
class CacheKey:
    def _to_cache_key_part(self, obj: Any) -> str:
        """Take an arbitrary value from the CacheKey and make a string out of it."""
        if is_dataclass(obj):
            obj = asdict(obj)
        if isinstance(obj, Mapping):
            return "_".join(self._to_cache_key_part(part) for part in sorted(obj.items()))
        if isinstance(obj, Sequence) and not isinstance(obj, (bytes, str)):
            return "_".join(self._to_cache_key_part(part) for part in obj)
        return repr(obj)

    def to_filename(self, version: int = 1, class_name: Optional[str] = None) -> str:
        """Get the cache filename."""
        def by_name(field):
            return field.name

        values = ":".join(
            field.name + "=" + self._to_cache_key_part(getattr(self, field.name))
            for field in sorted(fields(self), key=by_name)
        )
        value_repr = repr(values)
        sha = get_bytes_sha256(value_repr.encode("utf-8"))
        class_name = class_name or self.__class__.__name__
        return f"{class_name}_v{version}_{sha}.json"


_CacheKeyType = Type[CacheKey]


class _Cached:
    _cache_path_: ClassVar[Optional[pathlib.Path]]
    _cache_version_: ClassVar[int]
    CacheKey: ClassVar[_CacheKeyType]

    @classmethod
    def _get_cache_filename(cls, key: CacheKey):
        """Get the cache filename from the key."""
        filename = key.to_filename(class_name=cls.__name__, version=cls._cache_version_)
        return cls._cache_path_ / filename

    @classmethod
    def from_cache(cls: Type[T], key: CacheKey) -> Optional[T]:
        """Load the object based on its key from the whatrecord cache."""
        if cls is Cached:
            raise RuntimeError(f"Class {cls} is not intended to be saved/loaded")

        if cls._cache_path_ is None:
            return None

        try:
            with open(cls._get_cache_filename(key), "rb") as fp:
                serialized = json.load(fp)
        except FileNotFoundError:
            return None
        except json.JSONDecodeError:
            logger.debug("Failed to deserialize %s %s", cls, key, exc_info=True)
            return None

        try:
            return apischema.deserialize(cls, serialized)
        except Exception:
            logger.debug("Failed to deserialize %s %s", cls, key, exc_info=True)
            return None

    def save_to_cache(self, pretty: bool = False, overwrite: bool = True) -> bool:
        """Save the object to the whatrecord cache."""
        if self._cache_path_ is None:
            return False

        filename = self._get_cache_filename(self.key)
        if filename.exists() and not overwrite:
            return False

        serialized = apischema.serialize(self)
        dump_args = {"indent": 4} if pretty else {}
        try:
            with open(filename, "wt") as fp:
                json.dump(serialized, fp, sort_keys=True, **dump_args)
        except Exception:
            logger.exception(
                "Failed to save %s(%s) to the cache. (filename = %s)",
                self.__class__.__name__, self.key,
                filename
            )
            return False

        return True


@dataclass
class Cached(_Cached):
    """
    A generic dataclass that can be cached in ``WHATRECORD_CACHE_PATH``.

    Expects to be subclassed and configured with a CacheKey subclass.
    """

    key: CacheKey = field(metadata=apischema.metadata.skip)

    def __init_subclass__(
        cls,
        key: Optional[_CacheKeyType] = None,
        version: int = 1,
        cache_path: Optional[AnyPath] = None,
    ):
        if key is None or not inspect.isclass(key) or not is_dataclass(key):
            raise RuntimeError(
                f"{cls.__name__} should be defined with a keyword argument 'key'; "
                f"such as `class {cls.__name__}(Cached, key=CacheKeyClass):"
            )

        cls._cache_version_ = version
        cls.__annotations__["key"] = key
        cls.CacheKey = key
        if cache_path is None and not CACHE_PATH:
            cls._cache_path_ = None  # cache disabled
        else:
            cls._cache_path_ = pathlib.Path(cache_path or CACHE_PATH)


@dataclass
class InlineCached(_Cached):
    """
    A generic dataclass that can be cached in ``WHATRECORD_CACHE_PATH``.

    Expects to be subclassed and mixed in with a CacheKey subclass.
    """

    def __init_subclass__(
        cls,
        version: int = 1,
        cache_path: Optional[AnyPath] = None,
    ):
        for supercls in cls.mro()[1:]:
            if issubclass(supercls, CacheKey) and supercls is not CacheKey:
                cls.CacheKey = supercls
                break
        else:
            raise RuntimeError(
                f"{cls.__name__} should be defined as a subclass of a CacheKey"
            )

        cls._cache_version_ = version
        if cache_path is None and not CACHE_PATH:
            cls._cache_path_ = None  # cache disabled
        else:
            cls._cache_path_ = pathlib.Path(cache_path or CACHE_PATH)

    @property
    def key(self) -> CacheKey:
        """An auto-generated CacheKey based on this dataclass's attributes."""
        kwargs = {
            field.name: getattr(self, field.name)
            for field in fields(self.CacheKey)
        }
        return self.CacheKey(**kwargs)



import textwrap

import pytest

from .. import iocmanager


def test_parse_empty():
    assert iocmanager.parse_config([]) == []


@pytest.mark.parametrize(
    "source, expected",
    [
        pytest.param(
            textwrap.dedent(
                """\
        COMMITHOST = "psbuild-rhel7-01"

        hosts = [
           'amo-console',
           'ioc-amo-gige01',
           'ioc-amo-gige02',
           'ioc-amo-gige03',
           'ioc-amo-hv1',
           'ioc-amo-hv2',
           'ioc-amo-kbo-lvdt',
           'ioc-amo-lampenc01',
           'ioc-amo-las1',
           'ioc-amo-misc01',
           'ioc-amo-mot01',
           'ioc-amo-sdl',
           'ioc-amo-slits',
           'ioc-amo-uniq',
           'ioc-amo-vacuum',
           'ioc-det-pnccd01',
        ]

        procmgr_config = [
         {id:'ioc-amo-gige-01', host: 'ioc-amo-gige01', port: 30011, dir: 'ioc/amo/gigECam/R2.3.3',
          history: ['ioc/amo/gigECam/R2.3.3', '/reg/neh/home/sstubbs/work/amo/gigECam/current']},
         {id:'ioc-amo-gige-05', host: 'ioc-amo-gige01', port: 30015, dir: 'ioc/amo/gigECam/R2.3.3',
          history: ['ioc/amo/gigECam/R2.3.3']},
         {id:'ioc-amo-gige-06', host: 'ioc-amo-gige01', port: 30016, dir: 'ioc/amo/gigECam/R2.3.3',
          history: ['ioc/amo/gigECam/R2.3.3']},
         {id:'ioc-amo-gige-08', host: 'ioc-amo-gige01', port: 30029, dir: 'ioc/amo/gigECam/R2.3.3',
          history: ['ioc/amo/gigECam/R2.3.3']},
         {id:'ioc-amo-gige-09', host: 'ioc-amo-gige01', port: 30017, dir: 'ioc/amo/gigECam/R2.3.3',
          history: ['ioc/amo/gigECam/R2.3.3']},
         {id:'ioc-amo-usr-acromag', host: 'ioc-amo-gige01', port: 30001,
          dir: 'ioc/amo/pciAcromag/R1.0.4',
          history: ['ioc/amo/pciAcromag/R1.0.4']},
         {id:'ioc-amo-gige-02', host: 'ioc-amo-gige02', port: 30012, dir: 'ioc/amo/gigECam/R2.3.3',
          history: ['ioc/amo/gigECam/R2.3.3']},
         {id:'ioc-amo-gige-03', host: 'ioc-amo-gige02', port: 30013, dir: 'ioc/amo/gigECam/R2.3.3',
          history: ['ioc/amo/gigECam/R2.3.3']},
        ]
            """,
            ),
            [
                {
                    "id": "ioc-amo-gige-01",
                    "host": "ioc-amo-gige01",
                    "port": 30011,
                    "dir": "ioc/amo/gigECam/R2.3.3",
                    "history": [
                        "ioc/amo/gigECam/R2.3.3",
                        "/reg/neh/home/sstubbs/work/amo/gigECam/current",
                    ],
                },
                {
                    "id": "ioc-amo-gige-05",
                    "host": "ioc-amo-gige01",
                    "port": 30015,
                    "dir": "ioc/amo/gigECam/R2.3.3",
                    "history": ["ioc/amo/gigECam/R2.3.3"],
                },
                {
                    "id": "ioc-amo-gige-06",
                    "host": "ioc-amo-gige01",
                    "port": 30016,
                    "dir": "ioc/amo/gigECam/R2.3.3",
                    "history": ["ioc/amo/gigECam/R2.3.3"],
                },
                {
                    "id": "ioc-amo-gige-08",
                    "host": "ioc-amo-gige01",
                    "port": 30029,
                    "dir": "ioc/amo/gigECam/R2.3.3",
                    "history": ["ioc/amo/gigECam/R2.3.3"],
                },
                {
                    "id": "ioc-amo-gige-09",
                    "host": "ioc-amo-gige01",
                    "port": 30017,
                    "dir": "ioc/amo/gigECam/R2.3.3",
                    "history": ["ioc/amo/gigECam/R2.3.3"],
                },
                {
                    "id": "ioc-amo-usr-acromag",
                    "host": "ioc-amo-gige01",
                    "port": 30001,
                    "dir": "ioc/amo/pciAcromag/R1.0.4",
                    "history": ["ioc/amo/pciAcromag/R1.0.4"],
                },
                {
                    "id": "ioc-amo-gige-02",
                    "host": "ioc-amo-gige02",
                    "port": 30012,
                    "dir": "ioc/amo/gigECam/R2.3.3",
                    "history": ["ioc/amo/gigECam/R2.3.3"],
                },
                {
                    "id": "ioc-amo-gige-03",
                    "host": "ioc-amo-gige02",
                    "port": 30013,
                    "dir": "ioc/amo/gigECam/R2.3.3",
                    "history": ["ioc/amo/gigECam/R2.3.3"],
                },
            ],
            id="basic",
        ),
        pytest.param(
            textwrap.dedent(
                """\
                COMMITHOST = "psbuild-rhel7-01"

                hosts = [
                   'amo-console',
                   'ioc-amo-gige01',
                   'ioc-amo-gige02',
                   'ioc-amo-gige03',
                   'ioc-amo-hv1',
                   'ioc-amo-hv2',
                   'ioc-amo-kbo-lvdt',
                   'ioc-amo-lampenc01',
                   'ioc-amo-las1',
                   'ioc-amo-misc01',
                   'ioc-amo-mot01',
                   'ioc-amo-sdl',
                   'ioc-amo-slits',
                   'ioc-amo-uniq',
                   'ioc-amo-vacuum',
                   'ioc-det-pnccd01',
                ]

                procmgr_config = [
                     {id:'ioc-mec-gige11', host: 'ioc-mec-las-gige02',
                      port: 32511, dir: 'ioc/mec/gigECam/R1.1.1',
                      disable: True, alias: 'Gige Long Pulse 1',
                      history: ['ioc/mec/gigECam/R0.1.9']},
                    # {id:'ioc-mec-gige12', host: 'ioc-mec-las-gige02', port: 32512,
                    #  dir: 'ioc/mec/gigECam/R1.1.1', alias: 'Gige Long Pulse 2',
                    #  history: ['ioc/mec/gigECam/R0.1.9']},
                     {id:'ioc-mec-gige12', host: 'ioc-mec-las-gige02', port: 32512,
                      dir: '/cds/userpath', alias: 'Gige Long Pulse 2',
                      history: ['ioc/mec/gigECam/R0.1.9']},
                ]
            """,
            ),
            [
                {
                    "id": "ioc-mec-gige11",
                    "host": "ioc-mec-las-gige02",
                    "port": 32511,
                    "dir": "ioc/mec/gigECam/R1.1.1",
                    "disable": True,
                    "alias": "Gige Long Pulse 1",
                    "history": ["ioc/mec/gigECam/R0.1.9"],
                },
                {
                    "id": "ioc-mec-gige12",
                    "host": "ioc-mec-las-gige02",
                    "port": 32512,
                    "dir": "/cds/userpath",
                    "alias": "Gige Long Pulse 2",
                    "history": ["ioc/mec/gigECam/R0.1.9"],
                },
            ],
            id="commented",
        ),
    ]
)
def test_parse(source: str, expected: list, caplog: pytest.LogCaptureFixture):
    print(source)
    # print(source.splitlines())
    caplog.set_level("WARNING", logger=iocmanager.logger.name)
    assert iocmanager.parse_config(source.splitlines()) == expected
    assert not caplog.records, "No failure to parse messages"


@pytest.mark.parametrize(
    "source",
    [
        pytest.param(
            textwrap.dedent(
                """\
                procmgr_config=[
                 syntax error id:'ioc-amo-gige-01', host: 'ioc-amo-gige01',
                    port: 30011, dir: 'ioc/amo/gigECam/R2.3.3',
                  history: ['ioc/amo/gigECam/R2.3.3',
                            '/reg/neh/home/sstubbs/work/amo/gigECam/current']},
                ]
            """,
            ),
            id="syntax-error-1"
        ),
        pytest.param(
            textwrap.dedent(
                """\
                procmgr_config=[
                 {id:ioc-amo-gige-01', host: 'ioc-amo-gige01', port: 30011,
                  dir: 'ioc/amo/gigECam/R2.3.3',
                  history: ['ioc/amo/gigECam/R2.3.3',
                            '/reg/neh/home/sstubbs/work/amo/gigECam/current']},
                ]
            """,
            ),
            id="syntax-error-2"
        ),
        pytest.param(
            textwrap.dedent(
                """\
                procmgr_config=[
                 {id:ioc-amo-gige-01', host: 'ioc-amo-gige01', port: 30011,
                  dir: 'ioc/amo/gigECam/R2.3.3',
                  history; ['ioc/amo/gigECam/R2.3.3',
                            '/reg/neh/home/sstubbs/work/amo/gigECam/current']},
                ]
            """,
            ),
            id="syntax-error-3"
        ),
    ]
)
def test_parse_failure(source: str, caplog: pytest.LogCaptureFixture):
    print(source)
    # print(source.splitlines())
    caplog.set_level("WARNING", logger=iocmanager.logger.name)
    assert iocmanager.parse_config(source.splitlines()) == []
    assert len(caplog.records) == 1, "Failure to parse id line"



from __future__ import annotations

import logging
import math
import pathlib
import textwrap
import typing
from dataclasses import field
from typing import (Any, ClassVar, Dict, FrozenSet, Generator, List, Mapping,
                    Optional, Tuple, Union, cast)

import lark

from . import transformer, util
from .common import (DatabaseDevice, DatabaseMenu, LinterError, LinterMessage,
                     LinterWarning, LoadContext, PVAFieldReference,
                     RecordField, RecordInstance, RecordType, RecordTypeField,
                     StringWithContext, UnquotedString, dataclass)
from .macro import MacroContext
from .transformer import context_from_token

if typing.TYPE_CHECKING:
    from .shell import LoadedIoc, ShellState

logger = logging.getLogger(__name__)


def split_record_and_field(pvname) -> Tuple[str, str]:
    """Split REC.FLD into REC and FLD."""
    record, *field = pvname.split(".", 1)
    return record, field[0] if field else ""


class DatabaseLoadFailure(Exception):
    """Database load failure."""
    ...


@dataclass(repr=False)
class DatabaseLint:
    """
    Container for dbdlint results, with easier-to-access attributes.

    Reimplementation of ``pyPDB.dbdlint.Results``.

    Each error or warning has dictionary keys::

        {name, message, file, line, raw_message, format_args}

    Attributes
    ----------
    errors : list
        List of errors found
    warnings : list
        List of warnings found
    """
    errors: List[LinterError] = field(default_factory=list)
    warnings: List[LinterWarning] = field(default_factory=list)


@dataclass
class _TransformerState:
    """Transformer state for database parsing."""
    lint: DatabaseLint
    _record: Optional[RecordInstance] = None
    _record_type: Optional[RecordType] = None
    standalone_aliases: Dict[str, str] = field(default_factory=dict)
    aliases_to_update: List[Tuple[RecordInstance, str]] = field(
        default_factory=list
    )
    pva_references_to_update: List[Tuple[RecordInstance, PVAFieldReference]] = field(
        default_factory=list
    )
    # Record names aren't known until after all their fields are processed,
    # so use a magic token here for now.  If this leaks out, that's a bug!
    unset_name: str = "_whatrecord_processing_"

    def reset_record(self):
        """Reset the current record instance."""
        self._record = None

    @property
    def record(self) -> RecordInstance:
        """The current record instance."""
        if self._record is None:
            self._record = RecordInstance(
                context=(),
                name=self.unset_name,
                record_type="",
            )
        return self._record

    def reset_record_type(self):
        """Reset the current record type."""
        self._record_type = None

    @property
    def record_type(self) -> RecordType:
        """The current record type."""
        if self._record_type is None:
            self._record_type = RecordType(
                context=(),
                name="",
            )
        return self._record_type


@lark.visitors.v_args(inline=True)
class _DatabaseTransformer(lark.visitors.Transformer_InPlaceRecursive):
    def __init__(self, fn, dbd=None, enable_linting: bool = True):
        self.fn = str(fn)
        self.dbd = dbd
        self.db = Database()
        self._state = _TransformerState(lint=self.db.lint)

    @property
    def record_types(self) -> Dict[str, RecordType]:
        """
        Record types, either defined in an external .dbd file or in the one
        currently being loaded.
        """
        if self.dbd is not None:
            return self.dbd.record_types
        return self.db.record_types

    @lark.visitors.v_args(tree=True)
    def database(self, body) -> Database:
        """
        The final database generation step. The return value becomes the result
        of ``grammar.parse()``.
        """
        # Update all references that were set before we know the record name
        # Update record() { alias("") }
        for record, alias in self._state.aliases_to_update:
            self.db.aliases[alias] = record.name

        # Update pva Q:group field references
        for record, reference in self._state.pva_references_to_update:
            reference.record_name = record.name

        # Update top-level standalone aliases
        for alias_name, record_name in list(self._state.standalone_aliases.items()):
            self.db.standalone_aliases[alias_name] = record_name
            self.db.aliases[alias_name] = record_name
            if record_name in self.db.records:
                self.db.records[record_name].aliases.append(alias_name)
            # else:  linter error

        for device in self.db.devices:
            record_type = self.record_types.get(device.record_type, None)
            if record_type is not None:
                record_type.devices.append(device)

        return self.db

    dbitem = transformer.tuple_args

    def include(self, path):
        # raise RuntimeError("Incomplete dbd files are a TODO :(")
        self.db.includes.append(str(path))

    def path(self, path):
        self.db.paths.append(path)

    def addpath(self, path):
        self.db.addpaths.append(path)

    def menu(self, menu_token: lark.Token, name, choices):
        self.db.menus[name] = DatabaseMenu(
            context=context_from_token(self.fn, menu_token),
            name=name,
            choices=choices
        )

    menu_head = transformer.pass_through
    menu_body = transformer.dictify

    def device(
        self, _, record_type: str, link_type: str, dset_name: str, choice_string: str
    ):
        self.db.devices.append(
            DatabaseDevice(record_type, link_type, dset_name, choice_string)
        )

    def driver(self, _, drvet_name):
        self.db.drivers.append(drvet_name)

    def link(self, _, name, identifier):
        self.db.links[name] = identifier

    def registrar(self, _, name):
        self.db.registrars.append(name)

    def function(self, _, name):
        self.db.functions.append(name)

    def variable(self, _, name, dtype=None):
        self.db.variables[name] = dtype

    def breaktable(self, _, name, values):
        self.db.breaktables[name] = list(values)

    break_head = transformer.pass_through
    break_body = transformer.pass_through
    break_list = transformer.tuple_args
    break_item = transformer.pass_through

    def choice(self, _, identifier, string):
        return (identifier, string)

    def recordtype_field(self, field_tok, head, body):
        name, type_ = head
        known_keys = (
            "asl",
            "base",
            "extra",
            "initial",
            "interest",
            "menu",
            "pp",
            "prompt",
            "promptgroup",
            "prop",
            "size",
            "special",
        )
        kwargs = {
            key: body.pop(key)
            for key in known_keys
            if key in body
        }

        if body:
            # Only add this in if necessary; otherwise we can skip serializing
            # it.
            kwargs["body"] = body

        self._state.record_type.fields[name] = RecordTypeField(
            name=name,
            type=type_,
            context=context_from_token(self.fn, field_tok),
            **kwargs
        )

    recordtype_head = transformer.pass_through
    recordtype_field_body = transformer.dictify
    recordtype_field_head = transformer.tuple_args
    recordtype_field_item = transformer.tuple_args
    recordtype_field_item_menu = transformer.tuple_args

    def standalone_alias(self, _, record_name, alias_name):
        self._state.standalone_aliases[alias_name] = record_name

    def unquoted_string(self, value):
        return UnquotedString(value)

    def json_string(self, value):
        if value and value[0] in "'\"":
            return value[1:-1]
        return UnquotedString(value)

    string = json_string

    def json_key(self, value) -> StringWithContext:
        # Add context information for keys, especially for Q:group info nodes
        return StringWithContext(
            self.json_string(value),
            context=context_from_token(self.fn, value),
        )

    def json_array(self, elements=None):
        return elements or []

    def json_elements(self, *elements):
        return elements

    def JSON_TRUE(self, _):
        return True

    def JSON_FALSE(self, _):
        return False

    def JSON_NULL(self, _):
        return None

    def nan(self):
        return math.nan

    def hexint(self, sign, _, digits):
        return f"{sign}0x{digits}"

    def cdef(self, cdef_text):
        self._state.record_type.cdefs.append(str(cdef_text)[1:].strip())

    def recordtype(self, recordtype_token, name, body):
        record_type = self._state.record_type
        record_type.context = context_from_token(self.fn, recordtype_token)
        record_type.name = name
        self.db.record_types[name] = self._state.record_type
        self._state.reset_record_type()

    def record(self, rec_token, head, body):
        record_type, name = head
        record = self._state.record
        record.name = name
        record.context = context_from_token(self.fn, rec_token)
        record.is_grecord = rec_token == "grecord"
        record.is_pva = False
        record.record_type = record_type
        self.db.records[name] = self._state.record

        record_type_info = self.record_types.get(
            record.record_type, None
        )
        if record_type_info is None:
            # TODO lint error, if dbd loaded
            record.has_dbd_info = False
        else:
            record.has_dbd_info = True
            for fld in record.fields.values():
                field_info = record_type_info.fields.get(fld.name, None)
                if field_info is None:
                    # TODO lint error, if dbd loaded
                    ...
                else:
                    fld.dtype = field_info.type
                    # TODO: not 100% sure about the value of retaining the
                    # script name + line number in every field just yet;
                    # know for certain it's a waste of memory and repetetive
                    # data being sent over the wire at the very least
                    # fld.context = field_info.context[:1] + fld.context
        self._state.reset_record()

    json_dict = transformer.dictify
    json_key_value = transformer.tuple_args
    record_head = transformer.tuple_args

    def _add_lint(self, item: LinterMessage):
        """
        Track a new piece of lint.

        Parameters
        ----------
        item : LinterMessage
            The item to add to the DatabaseLint instance.
        """
        if isinstance(item, LinterWarning):
            self._state.lint.warnings.append(item)
        elif isinstance(item, LinterError):
            self._state.lint.errors.append(item)
        else:
            raise ValueError(f"Unexpected linter item: {item}")

    def record_field(self, field_token: lark.Token, name: str, value: Any):
        if isinstance(value, UnquotedString):
            self._add_lint(
                LinterWarning(
                    name="unquoted_field",
                    context=[LoadContext(self.fn, field_token.line)],
                    message=f"Unquoted field value {name!r}"
                ),
            )

        self._state.record.fields[name] = RecordField(
            dtype='', name=name, value=value,
            context=context_from_token(self.fn, field_token)
        )

    def _pva_q_group_handler(self, group: RecordInstance, md: Mapping):
        """Handler for Q:group."""

        # record(...) {
        #     info(Q:group, {      # <--- this thing
        #         "<group_name>":{
        #             +id:"some/NT:1.0",
        #             +meta:"FLD",
        #             +atomic:true,
        #             "<field.name>":{
        #                 +type:"scalar",
        #                 +channel:"VAL",
        #                 +id:"some/NT:1.0",
        #                 +trigger:"*",
        #                 +putorder:0,
        #             }
        #         }
        #     })
        # }
        for field_name, field_info in md.items():
            if not isinstance(field_info, dict):
                continue

            try:
                fieldref = group.fields[field_name]
            except KeyError:
                fieldref = group.fields[field_name] = PVAFieldReference(
                    name=str(field_name),
                    context=tuple(field_name.context),
                )

            # There, uh, is still some work left to do here.
            channel = field_info.get("+channel", None)
            if channel is not None:
                # The current record doesn't have its name yet due to how
                # the parser goes depth first; update it later.
                self._state.pva_references_to_update.append(
                    (self._state.record, fieldref)
                )
                fieldref.field_name = channel
                # Linter TODO: checks that this field exists and
                # whatnot

            fieldref.metadata.update(field_info)

    def _add_q_group(self, group_md: Mapping):
        """
        Handle qsrv "Q:" info nodes, and assemble new pseudo-record
        PVA ``RecordInstance`` out of them.
        """
        for group_name, group_info_to_add in group_md.items():
            if group_name not in self.db.pva_groups:
                self.db.pva_groups[group_name] = RecordInstance(
                    context=group_name.context,
                    name=str(group_name),
                    record_type="PVA",
                    is_pva=True,
                )

            self._pva_q_group_handler(
                self.db.pva_groups[group_name],
                group_info_to_add
            )

    def record_field_info(self, info_token: lark.Token, name: str, value: Any):
        record: RecordInstance = self._state.record
        context = context_from_token(self.fn, info_token)
        key = StringWithContext(name, context)
        record.info[key] = value

        if name == "Q:group" and isinstance(value, Mapping):
            self._add_q_group(value)

    def record_field_alias(self, _, name):
        record = self._state.record
        record.aliases.append(name)
        self._state.aliases_to_update.append((record, name))

    record_body = transformer.tuple_args


@dataclass
class Database:
    """
    Representation of an EPICS database, database definition, or both.

    Attributes
    ----------
    standalone_aliases : Dict[str, str]
        Standalone aliases are those defined outside of the record body; this
        may only be useful for faithfully reconstructing the Database according
        to its original source code.  Keyed on alias to actual record name.

    aliases : Dict[str, str]
        Alias name to record name.

    paths : List[str]
        The path command specifies the current search path for use when loading
        database and database definition files. The addpath appends directory
        names to the current path. The path is used to locate the initial
        database file and included files. An empty dir at the beginning,
        middle, or end of a non-empty path string means the current directory.

    addpaths : List[str]
        See 'paths' above.

    breaktables : Dict[str, List[str]]
        Breakpoint table (look-up table) of raw-to-engineering values.

    comments : List[str]
        Comments encountered while parsing the database.

    devices : List[DatabaseDevice]
        Device support declarations (dset).

    drivers : List[str]
        Driver declarations (drvet).

    functions : List[str]
        Exported C function names.

    includes : List[str]
        Inline inclusion. Not supported just yet.

    links : Dict[str, str]
        Links.

    menus : Dict[str, DatabaseMenu]
        Named value enumerations (enums).

    records : Dict[str, RecordInstance]
        Record name to RecordInstance.

    record_types : Dict[str, RecordType]
        Record type name to RecordType.

    registrars : List[str]
        Exported registrar function name.

    variables : Dict[str, Optional[str]]
        IOC shell variables.

    lint : DatabaseLint
        Any lint found when loading the database.
    """

    addpaths: List[str] = field(default_factory=list)
    aliases: Dict[str, str] = field(default_factory=dict)
    breaktables: Dict[str, List[str]] = field(default_factory=dict)
    comments: List[str] = field(default_factory=list)
    devices: List[DatabaseDevice] = field(default_factory=list)
    drivers: List[str] = field(default_factory=list)
    functions: List[str] = field(default_factory=list)
    includes: List[str] = field(default_factory=list)
    links: Dict[str, str] = field(default_factory=dict)
    menus: Dict[str, DatabaseMenu] = field(default_factory=dict)
    paths: List[str] = field(default_factory=list)
    pva_groups: Dict[str, RecordInstance] = field(default_factory=dict)
    record_types: Dict[str, RecordType] = field(default_factory=dict)
    records: Dict[str, RecordInstance] = field(default_factory=dict)
    registrars: List[str] = field(default_factory=list)
    standalone_aliases: Dict[str, str] = field(default_factory=dict)
    variables: Dict[str, Optional[str]] = field(default_factory=dict)

    lint: DatabaseLint = field(default_factory=DatabaseLint)

    _jinja_format_: ClassVar[Dict[str, str]] = {
        "file": textwrap.dedent(
            """\
            {% for include in includes %}
            include {{ include }}
            {% endfor %}
            {% for addpath in addpaths %}
            addpath {{ addpath }}
            {% endfor %}
            {% for path in paths %}
            path {{ path }}
            {% endfor %}
            {% for name, menu in menus.items() %}
            {{ render_object(menu, "file") }}
            {% endfor %}
            {% for name, record_type in record_types.items() %}
            {{ render_object(record_type, "file") }}
            {% for device in record_type.devices %}
            {{ render_object(device, "file") }}
            {% endfor %}
            {% endfor %}
            {% for link, identifier in links.items() %}
            link({{ link }}, {{ identifier }})
            {% endfor %}
            {% for driver in drivers %}
            driver({{ driver }})
            {% endfor %}
            {% for registrar in registrars %}
            registrar({{ registrar }})
            {% endfor %}
            {% for function in functions %}
            function({{ function }})
            {% endfor %}
            {% for variable, type in variables.items() %}
            {% if type %}
            variable({{ variable }}, {{ type }})
            {% else %}
            variable({{ variable }})
            {% endif %}
            {% endfor %}
            {% for name, breaktable in breaktables.items() %}
            breaktable({{ name }}) {
            {{ _indent }}{% for value in breaktable %}
            {{ _indent }}{{ value }}
            {{ _indent }}{% endfor %}
            }
            {% endfor %}
            {% for name, record in obj.non_aliased_records.items() %}
            {{ render_object(record, "file") }}

            {% endfor %}
            {% for alias, record in standalone_aliases.items() %}
            {% if record not in records or alias not in records[record].aliases %}
            alias("{{ alias }}", "{{ record }}")
            {% endif %}
            {% endfor %}
            """.rstrip()
        ),
    }

    @classmethod
    def from_string(
        cls,
        contents: str,
        dbd: Optional[Union[Database, str, pathlib.Path]] = None,
        filename: Optional[Union[str, pathlib.Path]] = None,
        macro_context: Optional[MacroContext] = None,
        version: int = 4,
        include_aliases: bool = True,
    ) -> Database:
        """
        Load a database [definition] from a string.

        Parameters
        ----------
        contents : str
            The contents of the database file.
        dbd : Union[Database, str, pathlib.Path], optional
            The database definition (.dbd) path, if applicable and available.
        filename : Union[str, pathlib.Path], optional
            The filename to associate with the contents.
        macro_context : MacroContext, optional
            A macro context to use for expanding macros while loading the
            database.
        version : int, optional
            The epics-base version to assume when loading.  Use 3 for R3.15 and under,
            4 for more recent versions.
        include_aliases : bool, optional
            Include aliases as top-level records.

        Returns
        -------
        Database
        """
        if dbd is not None and not isinstance(dbd, Database):
            dbd = Database.from_file(dbd, version=version)

        comments = []
        grammar = lark.Lark.open_from_package(
            "whatrecord",
            f"db.v{version}.lark",
            search_paths=("grammar", ),
            parser="lalr",
            lexer_callbacks={"COMMENT": comments.append},
            transformer=_DatabaseTransformer(filename, dbd=dbd),
            maybe_placeholders=False,
            # Per-user `gettempdir` caching of the LALR grammar analysis way of
            # passing ``True`` here:
            cache=True,
        )
        if macro_context is not None:
            contents = macro_context.expand_by_line(contents).rstrip() + "\n"

        db = cast(Database, grammar.parse(contents))
        db.comments = comments

        if include_aliases:
            for record_name, record in list(db.records.items()):
                for alias in record.aliases:
                    db.records[alias] = record

        return db

    @classmethod
    def from_file_obj(
        cls,
        fp,
        dbd: Optional[Union[Database, str, pathlib.Path]] = None,
        filename: Optional[Union[str, pathlib.Path]] = None,
        macro_context: Optional[MacroContext] = None,
        version: int = 4,
        include_aliases: bool = True,
    ) -> Database:
        """
        Load a database [definition] from a file object.

        Parameters
        ----------
        fp : file or file-like object
            The file object to read from.
        dbd : Union[Database, str, pathlib.Path], optional
            The database definition (.dbd) path, if applicable and available.
        filename : Union[str, pathlib.Path], optional
            The filename to associate with the contents.  Defaults to
            ``fp.name``, if available.
        macro_context : MacroContext, optional
            A macro context to use for expanding macros while loading the
            database.
        version : int, optional
            The epics-base version to assume when loading.  Use 3 for R3.15 and under,
            4 for more recent versions.
        include_aliases : bool, optional
            Include aliases as top-level records.

        Returns
        -------
        Database
        """
        return cls.from_string(
            fp.read(),
            filename=filename or getattr(fp, "name", None),
            dbd=dbd,
            macro_context=macro_context,
            version=version,
            include_aliases=include_aliases,
        )

    @classmethod
    def from_file(
        cls,
        fn: Union[str, pathlib.Path],
        dbd: Optional[Union[Database, str, pathlib.Path]] = None,
        macro_context: Optional[MacroContext] = None,
        version: int = 4,
        include_aliases: bool = True,
    ) -> Database:
        """
        Load a database [definition] from a filename.

        Parameters
        ----------
        fn : str or pathlib.Path
            The path to the database file.
        dbd : Union[Database, str, pathlib.Path], optional
            The database definition (.dbd) path, if applicable and available.
        macro_context : MacroContext, optional
            A macro context to use for expanding macros while loading the
            database.
        version : int, optional
            The epics-base version to assume when loading.  Use 3 for R3.15 and under,
            4 for more recent versions.
        include_aliases : bool, optional
            Include aliases as top-level records.

        Returns
        -------
        Database
        """
        with open(fn, "rt") as fp:
            return cls.from_string(
                fp.read(),
                filename=fn,
                dbd=dbd,
                macro_context=macro_context,
                version=version,
                include_aliases=include_aliases,
            )

    def field_names_by_type(
        self, field_types: List[str]
    ) -> Dict[str, FrozenSet[str]]:
        """
        Generate dictionary of record type to frozenset of field names.

        This can be used in scenarios where database definition files are
        unavailable and link information is requested.

        Parameters
        ----------
        field_types : list of str
            Field types to look for.
        """
        by_rtype = {}
        for rtype, info in sorted(self.record_types.items()):
            by_rtype[rtype] = frozenset(
                field.name
                for field in info.fields.values()
                if field.type in field_types
            )
        return by_rtype

    def add_or_update_record(self, record: RecordInstance):
        """
        Update (or add) records given a dictionary of records.
        """
        if record.is_pva:
            existing_record = self.pva_groups.get(record.name, None)
        else:
            existing_record = self.records.get(record.name, None)

        if not existing_record:
            self.records[record.name] = record
        else:
            existing_record.update(record)

    def shallow_copy(self) -> Database:
        """A shallow copy of the database with new dictionaries/lists."""

        return type(self)(
            addpaths=list(self.addpaths),
            aliases=dict(self.aliases),
            breaktables=dict(self.breaktables),
            comments=list(self.comments),
            devices=list(self.devices),
            drivers=list(self.drivers),
            functions=list(self.functions),
            includes=list(self.includes),
            links=dict(self.links),
            menus=dict(self.menus),
            paths=list(self.paths),
            pva_groups=dict(self.pva_groups),
            record_types=dict(self.record_types),
            records=dict(self.records),
            registrars=list(self.registrars),
            standalone_aliases=dict(self.standalone_aliases),
            variables=dict(self.variables),
            lint=DatabaseLint(
                warnings=list(self.lint.warnings),
                errors=list(self.lint.errors),
            )
        )

    def append(self, other: Database):
        """
        Append the other database, best-effort updating existing entries.

        This is not likely to do everything correctly (TODO).
        """
        for instance in other.records.values():
            self.add_or_update_record(instance)

        for instance in other.pva_groups.values():
            self.add_or_update_record(instance)

        def _update_list(this, other):
            this.extend([v for v in other if v not in this])

        _update_list(self.addpaths, other.addpaths)
        _update_list(self.comments, other.comments)
        _update_list(self.devices, other.devices)
        _update_list(self.drivers, other.drivers)
        _update_list(self.functions, other.functions)
        _update_list(self.includes, other.includes)
        _update_list(self.paths, other.paths)
        _update_list(self.registrars, other.registrars)
        self.aliases.update(other.aliases)
        self.breaktables.update(other.breaktables)
        self.links.update(other.links)
        self.menus.update(other.menus)
        self.record_types.update(other.record_types or {})
        self.standalone_aliases.update(other.standalone_aliases)
        self.variables.update(other.variables)

    @classmethod
    def from_multiple(cls, *items: _DatabaseSource) -> Database:
        """
        Create a Database instance from multiple sources, including:

        * Other Database instances
        * LoadedIoc
        * ShellState
        """
        from .shell import LoadedIoc, ShellState

        db = cls()

        for item in items:
            if isinstance(item, Database):
                db.append(item)
            elif isinstance(item, (LoadedIoc, ShellState)):
                state = item.shell_state if isinstance(item, LoadedIoc) else item
                new_records = list(state.database.values()) + list(state.pva_database.values())
                for record in new_records:
                    db.add_or_update_record(record)
                db.aliases.update(state.aliases)
            else:
                raise ValueError(f"Expected {_DatabaseSource}, got {type(item)}")

        return db

    @classmethod
    def from_vendored_dbd(cls, version: int = 3) -> Database:
        """
        Load the vendored database definition file from whatrecord.

        This is a good fallback when you have a database file without a
        corresponding database definition file.

        Parameters
        ----------
        version : int, optional
            Use the old V3 style or new V3 style database grammar by specifying
            3 or 4, respectively.  Defaults to 3.

        Returns
        -------
        db : Database
        """
        if version <= 3:
            return cls.from_file(
                util.MODULE_PATH / "tests" / "iocs/v3_softIoc.dbd",
                version=version,
            )
        return cls.from_file(
            util.MODULE_PATH / "tests" / "iocs" / "softIoc.dbd",
            version=version,
        )

    def get_links_for_record(
        self,
        record: RecordInstance,
    ) -> Generator[Tuple[RecordField, str, List[str]], None, None]:
        """
        Get all links - in, out, and forward links.

        Parameters
        ----------
        record : RecordInstance
            Additional information, if the database definition wasn't loaded
            with this instance.

        Yields
        ------
        field : RecordField
        link_text: str
        link_info: str
        """
        record_info = self.record_types.get(record.record_type, None)
        if not record_info:
            return

        yield from record_info.get_links_for_record(record)

    @property
    def all_aliases(self) -> Dict[str, str]:
        """All aliases: top-level-defined and per-instance-defined."""
        aliases = dict(self.aliases)
        for record_name, record in self.records.items():
            for alias in record.aliases:
                aliases[alias] = record_name
        return aliases

    @property
    def non_aliased_records(self) -> Dict[str, RecordInstance]:
        """
        Get unique and non-aliased record instances.

        This can be used to ignore records included by the ``include_aliases``
        setting.
        """
        records = {}
        aliases = self.all_aliases
        for record_name, record in self.records.items():
            if record_name not in aliases:
                records[record_name] = record
        return records


_DatabaseSource = Union["LoadedIoc", "ShellState", Database]



import dataclasses
import functools
import inspect
import json
import logging
import textwrap
import typing
from typing import Optional

import apischema
import jinja2

from . import settings
from .common import LoadContext

logger = logging.getLogger(__name__)


pass_eval_context = (
    jinja2.pass_eval_context
    if hasattr(jinja2, "pass_eval_context")
    else jinja2.evalcontextfilter
)


@dataclasses.dataclass
class FormatOptions:
    indent: int = settings.INDENT


class FormatContext:
    format_options: FormatOptions

    def __init__(
        self,
        helpers: Optional[list] = None,
        options: Optional[FormatOptions] = None,
        *,
        trim_blocks: bool = True,
        lstrip_blocks: bool = False,
        default_options: str = "console",
        **env_kwargs
    ):
        self.helpers = helpers or [self.render_object, type, locals, json]
        self.default_options = default_options
        self._template_dict = {}
        self.env = jinja2.Environment(
            loader=jinja2.DictLoader(self._template_dict),
            trim_blocks=trim_blocks,
            lstrip_blocks=lstrip_blocks,
            **env_kwargs,
        )

        self.format_options = options or FormatOptions()
        self.env.filters.update(self.get_filters())
        self.default_render_context = self.get_render_context()
        self._fallback_formats = {}

    def get_filters(self, **user_config):
        """All jinja filters."""

        @pass_eval_context
        def title_fill(eval_ctx, text, fill_char):
            return fill_char * len(text)

        @pass_eval_context
        def classname(eval_ctx, obj):
            if inspect.isclass(obj):
                return obj.__name__
            return type(obj).__name__

        @pass_eval_context
        def render_object(eval_ctx, obj, option):
            return self.render_object(obj, option)

        return {
            key: value
            for key, value in locals().items()
            if not key.startswith("_") and key not in {"self"}
        }

    def render_template(self, _template: str, **context):
        # TODO: want this to be positional-only; fallback here for pypy
        template = _template

        for key, value in self.default_render_context.items():
            context.setdefault(key, value)
        context["render_ctx"] = context
        return self.env.from_string(template).render(context)

    def _render_object_fallback(self, _obj, _option, **context):
        # TODO: want this to be positional-only; fallback here for pypy
        obj, option = _obj, _option

        def render_prefixed(sub_obj, prefix: str) -> str:
            return textwrap.indent(
                self.render_object(sub_obj, option, **context),
                prefix=prefix,
            )

        if isinstance(obj, typing.Sequence) and not isinstance(obj, str):
            if all(isinstance(obj_idx, LoadContext) for obj_idx in obj):
                # Special-case FullLoadContext
                return " ".join(str(ctx) for ctx in obj)

            return "\n".join(
                f"[{idx}]: " + render_prefixed(obj_idx, '    ').lstrip()
                for idx, obj_idx in enumerate(obj)
            )

        if isinstance(obj, typing.Mapping):
            return "\n".join(
                f'"{key}": ' + render_prefixed(value, " " * (len(key) + 4))
                for key, value in sorted(obj.items())
            )

        if dataclasses.is_dataclass(obj):
            serialized = apischema.serialize(obj)
            return json.dumps(serialized, indent=self.format_options.indent)

        return str(obj)

    def render_object(self, _obj, _option=None, **context):
        # TODO: want this to be positional-only; fallback here for pypy
        obj, option = _obj, _option
        if option is None:
            option = self.default_options

        if dataclasses.is_dataclass(obj):
            for field in dataclasses.fields(obj):
                context.setdefault(field.name, getattr(obj, field.name))

        context.setdefault("obj", obj)

        try:
            template = obj._jinja_format_[option]
        except (AttributeError, KeyError):
            ...
        else:
            return self.render_template(template, **context)

        return self._render_object_fallback(obj, option, **context)

    def get_render_context(self) -> dict:
        """Jinja template context dictionary - helper functions."""
        context = {func.__name__: func for func in self.helpers}
        context["_fmt"] = self.format_options
        context["_json_dump"] = functools.partial(json.dumps, indent=self.format_options.indent)
        context["_indent"] = self.format_options.indent * " "
        return context



"""
LCLS-specific epicsArch.txt plugin.
"""

from __future__ import annotations

import argparse
import json
import logging
import pathlib
import typing
from dataclasses import dataclass, field
from typing import Dict, List, Optional, Type

import apischema
import lark

from .. import transformer
from ..common import AnyPath, FullLoadContext, LoadContext, StringWithContext
from ..db import split_record_and_field
from ..server.common import PluginResults
from ..util import get_bytes_sha256, get_file_sha256

logger = logging.getLogger(__name__)

# Stash the description for later usage by the CLI interface.
DESCRIPTION = __doc__.strip()


@dataclass
class LclsEpicsArchPluginResults(PluginResults):
    # Could potentially further specify metadata_by_key or metadata
    ...


@dataclass
class Comment:
    context: FullLoadContext
    text: str


@dataclass
class Warning:
    """epicsArch-format warning."""
    context: FullLoadContext
    type_: str
    text: str


@dataclass
class DaqPV:
    """A single PV configured for storage in the DAQ/logbook."""
    context: FullLoadContext
    name: str
    alias: str = ""
    provider: str = "ca"
    comments: List[Comment] = field(default_factory=list)


@dataclass
class LclsEpicsArchFile:
    """Representation of an LCLS-specific DAQ recording epicsArch.txt file."""
    pvs: Dict[str, DaqPV]
    aliases: Dict[str, str]
    warnings: List[Warning]
    filename: Optional[pathlib.Path] = None
    loaded_files: Dict[str, str] = field(default_factory=dict)

    @classmethod
    def from_string(
        cls,
        contents: str,
        filename: Optional[AnyPath] = None,
        debug: bool = False,
        context: Optional[FullLoadContext] = None,
    ) -> LclsEpicsArchFile:
        """Load an epicsArch.txt file given its string contents."""
        if filename:
            filename = pathlib.Path(filename).resolve()

        grammar = lark.Lark.open_from_package(
            "whatrecord",
            "lcls_epicsarch.lark",
            search_paths=("grammar",),
            parser="earley",
            propagate_positions=True,
            debug=debug,
        )

        transformer_ = _EpicsArchTransformer(
            cls, filename, contents, grammar, context=context
        )
        return transformer_.transform(grammar.parse(contents + "\n"))

    @classmethod
    def from_file_obj(
        cls, fp, filename: Optional[AnyPath] = None,
        context: Optional[FullLoadContext] = None,
    ) -> LclsEpicsArchFile:
        """Load an epicsArch.txt file from a file object."""
        return cls.from_string(
            fp.read(),
            filename=getattr(fp, "name", filename),
            context=context,
        )

    @classmethod
    def from_file(
        cls, fn: AnyPath,
        context: Optional[FullLoadContext] = None,
    ) -> LclsEpicsArchFile:
        """
        Load an epicsArch.txt file.

        Parameters
        ----------
        filename : pathlib.Path or str
            The filename.

        Returns
        -------
        file : LclsEpicsArchFile
            The parsed file.
        """
        with open(fn, "rt") as fp:
            return cls.from_string(fp.read(), filename=fn, context=context)


@lark.visitors.v_args(inline=True)
class _EpicsArchTransformer(lark.visitors.Transformer):
    cls: Type[LclsEpicsArchFile]
    fn: str
    _load_context: FullLoadContext
    _loaded_files: Dict[str, str]
    _path: pathlib.Path
    _pvs: Dict[str, DaqPV]
    _aliases: Dict[str, DaqPV]
    _comments: List[Comment]
    _warnings: List[Warning]
    _grammar: lark.Grammar
    _description: Optional[StringWithContext]
    _description_index: int

    def __init__(
        self,
        cls,
        fn: Optional[pathlib.Path],
        raw_contents: str,
        grammar: lark.Grammar,
        context: Optional[FullLoadContext] = None,
        visit_tokens=False,
    ):
        super().__init__(visit_tokens=visit_tokens)
        self.fn = str(fn)
        self.cls = cls
        self._loaded_files = {}
        self._load_context = context or ()
        self._aliases = {}
        self._comments = []
        self._grammar = grammar
        self._path = fn.resolve() if fn else pathlib.Path(".")
        self._pvs = {}
        self._warnings = []
        self._description = None
        self._description_index = 0

        if not fn:
            return

        self._loaded_files[str(fn)] = get_bytes_sha256(
            raw_contents.encode("utf-8")
        )

    def archfile(self, *entries):
        return self.cls(
            pvs=self._pvs,
            filename=pathlib.Path(self.fn).resolve(),
            warnings=self._warnings,
            aliases=self._aliases,
            loaded_files=self._loaded_files,
        )

    pvname = transformer.pass_through
    comment_text = transformer.pass_through

    def comment(self, comment_token, comment=None, *_):
        self._comments.append(
            Comment(
                context=transformer.context_from_token(self.fn, comment_token),
                text=str(comment).strip().lstrip("# \t"),
            )
        )

    def pv(self, pvname, *extra):
        """pvname (_WS+ provider)? EOL"""
        if len(extra) > 1:
            provider, *_ = extra
        else:
            provider = "ca"

        pv = DaqPV(
            context=transformer.context_from_token(self.fn, pvname),
            name=str(pvname),
            provider=str(provider),
        )

        if self._description:
            self._description_index += 1
            if self._description_index > 1:
                # NOTE: the old daq docs indicate it appended "-{pv}" on the
                # description... but the new source does this:
                pv.alias = f"{self._description}_{self._description_index}"
            else:
                pv.alias = self._description

        pv.comments = list(self._comments)
        self._comments = []
        self._add_pv(pv)
        return pv

    def description(self, desc_prefix, desc_text, _):
        """DESC_PREFIX _WS* DESC_TEXT EOL"""
        self._description = StringWithContext(
            desc_text,
            context=transformer.context_from_token(self.fn, desc_prefix)
        )
        self._description_index = 0

    pvs = transformer.tuple_args

    def _add_pv(self, pv: DaqPV):
        def combine_contexts(*contexts: FullLoadContext) -> FullLoadContext:
            result = []
            for ctx in contexts:
                for part in ctx:
                    if part not in result:
                        result.append(part)
            return result

        if pv.name in self._pvs:
            old_pv = self._pvs[pv.name]
            self._warnings.append(
                Warning(
                    context=combine_contexts(pv.context, old_pv.context),
                    type_="duplicate_pv",
                    text=f"Duplicate pvname: {pv.name} {old_pv.context}"
                )
            )

        if pv.alias in self._aliases:
            old_pv = self._pvs[self._aliases[pv.alias]]
            self._warnings.append(
                Warning(
                    context=combine_contexts(pv.context, old_pv.context),
                    type_="duplicate_alias",
                    text=f"Duplicate alias: {pv.alias} {old_pv.context}",
                )
            )

        if pv.alias in self._pvs:
            old_pv = self._pvs[pv.alias]
            self._warnings.append(
                Warning(
                    context=combine_contexts(pv.context, old_pv.context),
                    type_="alias_is_pv",
                    text=f"Alias name is a PV: {pv.alias} {old_pv.context}",
                )
            )

        if pv.name in self._aliases:
            old_pv = self._pvs[self._aliases[pv.name]]
            self._warnings.append(
                Warning(
                    context=combine_contexts(pv.context, old_pv.context),
                    type_="pv_is_alias",
                    text=f"PV name matches alias: {pv.name} {old_pv.context}",
                )
            )

        if pv.name not in self._pvs:
            self._pvs[pv.name] = pv
            # Tack on our load context
            pv.context = self._load_context[-1:] + pv.context

        if pv.alias:
            self._aliases[pv.alias] = pv.name

    def filenames(self, *filenames):
        return filenames

    def include(self, include_token, filenames, _):
        include_ctx = transformer.context_from_token(self.fn, include_token)
        for filename in filenames:
            filename = (self._path.parent / filename).resolve()
            if (
                any(ctx.name == str(filename) for ctx in self._load_context)
                or filename == self._path
            ):
                self._warnings.append(
                    Warning(
                        context=include_ctx,
                        type_="recursive_include",
                        text=(
                            f"Recursively included {filename} at "
                            f"{include_ctx[0]} {self._load_context}"
                        ),
                    )
                )
                continue

            try:
                included = self.cls.from_file(
                    filename,
                    context=self._load_context + include_ctx,
                )
            except lark.exceptions.LarkError as ex:
                line = getattr(ex, "line", 0)
                if not isinstance(line, int):
                    line = 0
                self._warnings.append(
                    Warning(
                        context=include_ctx + (LoadContext(str(filename), line), ),
                        type_="bad_file",
                        text=(
                            f"Failed to parse {filename}: {ex.__class__.__name__} {ex}"
                        )
                    )
                )
                try:
                    self._loaded_files[str(filename)] = get_file_sha256(filename)
                except Exception:
                    ...
            except FileNotFoundError:
                self._warnings.append(
                    Warning(
                        context=include_ctx,
                        type_="missing_file",
                        text=f"Included {filename} missing"
                    )
                )
            else:
                self._warnings.extend(included.warnings)
                for pv in included.pvs.values():
                    self._add_pv(pv)
                self._loaded_files.update(included.loaded_files)

        def blank_line(self, *_):
            self._description = None


def main(
    filenames: List[str],
    pretty: bool = False,
):
    by_filename = {}
    record_to_file = {}
    execution_info = {}
    files_to_monitor = {}
    for filename in sorted(filenames):
        try:
            archfile = LclsEpicsArchFile.from_file(filename)
        except Exception as ex:
            archfile = LclsEpicsArchFile(
                pvs={},
                aliases={},
                filename=pathlib.Path(filename),
                warnings=[
                    Warning(
                        context=(LoadContext(str(filename), 0), ),
                        type_="bad_file",
                        text=(
                            f"Failed to parse {filename}: {ex.__class__.__name__} {ex}"
                        )
                    )
                ],
            )

        files_to_monitor.update(archfile.loaded_files)
        filename = str(filename)
        by_filename[filename] = archfile
        for pv in sorted(archfile.pvs):
            record, _ = split_record_and_field(pv)
            record_to_file.setdefault(record, []).append(filename)

    return LclsEpicsArchPluginResults(
        files_to_monitor=files_to_monitor,
        record_to_metadata_keys=record_to_file,
        metadata={},
        metadata_by_key=by_filename,
        execution_info=execution_info,
    )


def _get_argparser(
    parser: typing.Optional[argparse.ArgumentParser] = None,
) -> argparse.ArgumentParser:
    if parser is None:
        parser = argparse.ArgumentParser(description=DESCRIPTION)

    pds_root = pathlib.Path("/cds/group/pcds/dist/pds")
    if pds_root.exists():
        default_files = (
            tuple(pds_root.glob("*/misc/epicsArch.txt")) +
            tuple(pds_root.glob("*/misc/logbook.txt"))
        )
    else:
        default_files = ()

    parser.add_argument(
        "-f", "--filename",
        dest="filenames",
        type=str,
        nargs="+",
        required=False,
        help="Filenames",
        default=default_files,
    )

    parser.add_argument(
        "-p", "--pretty", action="store_true", help="Pretty JSON output"
    )
    return parser


def _cli_main():
    parser = _get_argparser()
    kwargs = vars(parser.parse_args())
    results = main(**kwargs)
    json_results = apischema.serialize(results)
    dump_args = {"indent": 4} if kwargs["pretty"] else {}
    print(json.dumps(json_results, sort_keys=True, **dump_args))


if __name__ == "__main__":
    _cli_main()



import ast
import dataclasses
import os
import re
from typing import Any, Dict, Optional

import apischema
from epicsmacrolib import MacroContext

from . import settings


def macros_from_string(
    macro_string: str, use_environment: bool = False
) -> Dict[str, str]:
    """
    Get a macro dictionary from a macro string.

    Parameters
    ----------
    macro_string : str
        The macro string, in the format A=B,C=D,...

    use_environment : bool, optional
        Use environment variables as well.  Defaults to False.

    Returns
    -------
    macros : Dict[str, str]
        Macro key to value.
    """
    if not macro_string.strip():
        return {}
    macro_context = MacroContext(use_environment=use_environment)
    return macro_context.define_from_string(macro_string)


class PassthroughMacroContext(MacroContext):
    """
    A stand-in MacroContext which performs no macro expansion.

    That is, a pass-through/no-operation macro context.
    """

    def expand(self, value: str, **_) -> str:
        return value

    def expand_by_line(self, contents: str, *, delimiter: str = "\n") -> str:
        return delimiter.join(
            line
            for line in contents.splitlines()
        )


@dataclasses.dataclass
class _SerializedMacroContext:
    #: Show warnings
    show_warnings: bool
    #: String encoding to use internally with macLib.
    string_encoding: str
    #: The macros, including any environment variables (if use_environment set).
    macros: Dict[str, str]
    #: If the context is a passthrough one.
    passthrough: bool


RE_MACRO_KEY_SKIP = []

# This becomes more of a concern when run on CI.
# Consider tweaking this for your purposes in MACRO_KEY_SKIP or
# setting MACRO_NO_ENV for CI jobs.
MACRO_KEY_SKIP_DEFAULT = [
    ".*TOKEN.*",
    ".*SECRET.*",
    ".*SSH.*",
    ".*GITHUB.*",
]


def set_serialization_settings(skip: Optional[str] = settings.MACRO_KEY_SKIP):
    """Update macro serialization settings."""
    global RE_MACRO_KEY_SKIP
    if skip:
        skip_regex = ast.literal_eval(skip)
    else:
        skip_regex = MACRO_KEY_SKIP_DEFAULT
    RE_MACRO_KEY_SKIP = [re.compile(regex) for regex in skip_regex]


set_serialization_settings()


def should_serialize_key(key: str, value: str) -> bool:
    """
    Should ``key`` be serialized when saving macros?

    Settings used:
    * WHATRECORD_MACRO_KEY_SKIP (str) - keys to skip, specified by regular
      expressions.  Should be valid Python syntax for a list of strings, as it
      will be handed to ``ast.literal_eval``.
    * WHATRECORD_MACRO_INCLUDE_ENV (bool) - serialize environment variables if
      set to 1.
    * WHATRECORD_MACRO_VALUE_MAX_LENGTH (int) - maximum length of macro values
      to serialize. 0 to disable maximum length check.

    Parameters
    ----------
    key : str
        The macro/environment variable name.
    value : str
        The value associated with the macro.

    Returns
    -------
    bool
    """
    # Above the maximum length threshold -> skip
    if settings.MACRO_VALUE_MAX_LENGTH > 0:
        if len(value) > settings.MACRO_VALUE_MAX_LENGTH:
            return False

    # Throw out environment variables if MACRO_INCLUDE_ENV is unset:
    if not settings.MACRO_INCLUDE_ENV and key in os.environ:
        return False

    # Bad key or one that matches regular expressions in RE_MACRO_KEY_SKIP
    # (by way of 'WHATRECORD_MACRO_KEY_SKIP')
    if not key:
        return False
    return not any(regex.fullmatch(key) for regex in RE_MACRO_KEY_SKIP)


@apischema.serializer
def _serialize_macro_context(ctx: MacroContext) -> Dict[str, Any]:
    macros = {
        key: value
        for key, value in ctx.items()
        if should_serialize_key(key, value)
    }

    return apischema.serialize(
        _SerializedMacroContext(
            show_warnings=ctx.show_warnings,
            string_encoding=ctx.string_encoding,
            macros=macros,
            passthrough=isinstance(ctx, PassthroughMacroContext),
        )
    )


@apischema.deserializer
def _deserialize_macro_context(info: Dict[str, Any]) -> MacroContext:
    obj = apischema.deserialize(
        _SerializedMacroContext,
        info
    )

    cls = PassthroughMacroContext if obj.passthrough else MacroContext
    return cls(
        show_warnings=obj.show_warnings,
        string_encoding=obj.string_encoding,
        use_environment=False,
        macros=obj.macros,
    )


__all__ = ["MacroContext", "PassthroughMacroContext", "macros_from_string"]



"""V3 database parsing tests."""

import textwrap
from typing import Optional

import apischema
import pytest

from .. import Database, common
from ..common import LoadContext
from ..db import (DatabaseMenu, LinterWarning, RecordField, RecordInstance,
                  RecordType, RecordTypeField)
from ..format import FormatContext, FormatOptions

v3_or_v4 = pytest.mark.parametrize("version", [3, 4])


def test_simple():
    db = Database.from_string(
        """\
record(ai, "rec:X") {
    field(A, "test")
    field(B, test)
}
""",
        version=3
    )
    assert db.records["rec:X"] == RecordInstance(
        context=(LoadContext("None", 1), ),
        record_type="ai",
        name="rec:X",
        is_pva=False,
        fields={
            "A": RecordField(
                context=(LoadContext("None", 2), ),
                name="A",
                dtype="",
                value="test",
            ),
            "B": RecordField(
                context=(LoadContext("None", 3), ),
                name="B",
                dtype="",
                value="test",
            ),
        }
    )
    apischema.deserialize(Database, apischema.serialize(db))


def test_tab_in_field():
    db = Database.from_string(
        """\
record(ai, "rec:X") {
    field(A, "test\tvalue")
}
""",
        version=3
    )
    assert db.records["rec:X"] == RecordInstance(
        context=(LoadContext("None", 1), ),
        record_type="ai",
        name="rec:X",
        is_pva=False,
        fields={
            "A": RecordField(
                context=(LoadContext("None", 2), ),
                name="A",
                dtype="",
                value="test\tvalue",
            ),
        }
    )
    apischema.deserialize(Database, apischema.serialize(db))


@v3_or_v4
def test_breaktable(version):
    db = Database.from_string(
        """\
        breaktable(typeAttenLength) {
0.8      0.18
0.9      0.25
8.0    150.13
8.5    174.81
9.0    204.32
}
""",
        version=version
    )
    assert db.breaktables["typeAttenLength"] == [
        '0.8', '0.18', '0.9', '0.25', '8.0', '150.13', '8.5', '174.81', '9.0',
        '204.32'
    ]
    apischema.deserialize(Database, apischema.serialize(db))


@v3_or_v4
def test_dbd_recordtype(version):
    db = Database.from_string(
        """\
recordtype(stringin) {
    field(NAME, DBF_STRING) {
        size(61)
        special(SPC_NOMOD)
        prompt("Record Name")
    }
    field(PINI, DBF_MENU) {
        menu(menuPini)
        interest(1)
        promptgroup("20 - Scan")
        prompt("Process at iocInit")
    }
}
""",
        version=version
    )
    assert db.record_types["stringin"] == RecordType(
        context=(LoadContext("None", 1), ),
        name="stringin",
        cdefs=[],
        fields={
            "NAME": RecordTypeField(
                context=(LoadContext("None", 2), ),
                name="NAME",
                type="DBF_STRING",
                special="SPC_NOMOD",
                prompt="Record Name",
                size="61",
                body={},
            ),
            "PINI": RecordTypeField(
                context=(LoadContext("None", 7), ),
                name="PINI",
                type="DBF_MENU",
                menu="menuPini",
                interest="1",
                promptgroup="20 - Scan",
                prompt="Process at iocInit",
                body={},
            ),
        }
    )
    apischema.deserialize(Database, apischema.serialize(db))


@v3_or_v4
def test_dbd_menus(version):
    db = Database.from_string(
        """\
menu(stringoutPOST) {
    choice(stringoutPOST_OnChange, "On Change")
    choice(stringoutPOST_Always, "Always")
}
menu(menuScan) {
    choice(menuScanPassive, "Passive")
    choice(menuScanEvent, "Event")
    choice(menuScanI_O_Intr, "I/O Intr")
    choice(menuScan10_second, "10 second")
    choice(menuScan5_second, "5 second")
    choice(menuScan2_second, "2 second")
    choice(menuScan1_second, "1 second")
    choice(menuScan_5_second, ".5 second")
    choice(menuScan_2_second, ".2 second")
    choice(menuScan_1_second, ".1 second")
}
""",
        version=version
    )
    assert db.menus["stringoutPOST"] == DatabaseMenu(
        context=(LoadContext("None", 1), ),
        name="stringoutPOST",
        choices={
            "stringoutPOST_OnChange": "On Change",
            "stringoutPOST_Always": "Always",
        },
    )

    assert db.menus["menuScan"] == DatabaseMenu(
        context=(LoadContext("None", 5), ),
        name="menuScan",
        choices={
            "menuScanPassive": "Passive",
            "menuScanEvent": "Event",
            "menuScanI_O_Intr": "I/O Intr",
            "menuScan10_second": "10 second",
            "menuScan5_second": "5 second",
            "menuScan2_second": "2 second",
            "menuScan1_second": "1 second",
            "menuScan_5_second": ".5 second",
            "menuScan_2_second": ".2 second",
            "menuScan_1_second": ".1 second",
        },
    )
    apischema.deserialize(Database, apischema.serialize(db))


@v3_or_v4
def test_dbd_cdef(version):
    db = Database.from_string(
        """\
recordtype(stringin) {
    %#include "test.h"
    %#include "test1.h"
    %#include "test2.h"
}
""",
        version=version
    )
    assert db.record_types["stringin"] == RecordType(
        context=(LoadContext("None", 1), ),
        name="stringin",
        cdefs=[
            '#include "test.h"',
            '#include "test1.h"',
            '#include "test2.h"',
        ],
        fields={},
    )
    apischema.deserialize(Database, apischema.serialize(db))


def test_alias_and_standalone_alias():
    db = Database.from_string(
        """\
record(ai, "rec:X") {
    alias("rec:Y")
    field(A, "test")
    field(B, test)
}
alias("rec:X", "rec:Z")
""",
        version=3
    )
    assert db.aliases["rec:Y"] == "rec:X"
    assert db.aliases["rec:Z"] == "rec:X"
    assert db.standalone_aliases["rec:Z"] == "rec:X"
    assert db.records["rec:X"] == RecordInstance(
        context=(LoadContext("None", 1), ),
        record_type="ai",
        name="rec:X",
        is_pva=False,
        aliases=["rec:Y", "rec:Z"],
        fields={
            "A": RecordField(
                context=(LoadContext("None", 3), ),
                name="A",
                dtype="",
                value="test",
            ),
            "B": RecordField(
                context=(LoadContext("None", 4), ),
                name="B",
                dtype="",
                value="test",
            ),
        }
    )
    apischema.deserialize(Database, apischema.serialize(db))


def test_standalone_alias():
    db = Database.from_string(
        """\
record(ai, "rec:X") {
    field(A, "test")
    field(B, test)
}
alias("rec:X", "rec:Y")
""",
        version=3
    )
    assert db.records["rec:X"] == RecordInstance(
        context=(LoadContext("None", 1), ),
        record_type="ai",
        name="rec:X",
        is_pva=False,
        aliases=["rec:Y"],
        fields={
            "A": RecordField(
                context=(LoadContext("None", 2), ),
                name="A",
                dtype="",
                value="test",
            ),
            "B": RecordField(
                context=(LoadContext("None", 3), ),
                name="B",
                dtype="",
                value="test",
            ),
        }
    )
    apischema.deserialize(Database, apischema.serialize(db))


def test_unquoted_warning():
    db = Database.from_string(
        """\
record(ai, "rec:X") {
    field(A, "test")
    field(B, test)
}
""",
        version=3
    )
    assert db.lint.warnings == [
        LinterWarning(
            name="unquoted_field",
            context=[LoadContext("None", 3)],
            message="Unquoted field value 'B'"
        )
    ]


@pytest.mark.parametrize(
    "version", [3, 4],
)
def test_load_vendored_database_smoke(version: int):
    dbd = Database.from_vendored_dbd(version=version)
    record_types = list(dbd.record_types.values())
    assert len(record_types)
    record = record_types[0]
    if version == 3:
        assert "v3_softIoc.dbd" in record.context[0].name
    else:
        assert "softIoc.dbd" in record.context[0].name


@pytest.mark.parametrize(
    "base_version, expected",
    [
        ("3", 3),
        ("3.14.12.2", 3),
        ("3.14", 3),
        ("3.15", 3),
        ("3.16", 4),
        ("4.0", 4),
        ("7.0", 4),
        ("7.0.3.1", 4),
    ],
)
def test_grammar_version_check(base_version: str, expected: int):
    assert common.get_grammar_version_by_base_version(base_version) == expected


@pytest.mark.parametrize(
    "input_text, expected",
    [
        pytest.param(
            """\
            menu(stringoutPOST) {
                choice(stringoutPOST_OnChange, "On Change")
                choice(stringoutPOST_Always, "Always")
            }
            recordtype(ai) {
                %#include "epicsTypes.h"
                %#include "link.h"
                field(NAME, DBF_STRING) {
                    prompt("Record Name")
                    special(SPC_NOMOD)
                    size(61)
                }
            }
            device(ai, CONSTANT, devAaiSoft, "Soft Channel")
            registrar(rsrvRegistrar)
            variable(dbBptNotMonotonic, int)
            """,
            None,
            id="dbd_basic"
        ),
        pytest.param(
            """\
            record(ai, "rec:X") {
                field(A, "test")
                field(B, "test")
                info(info, "X")
            }
            """,
            None,
            id="record_basic"
        ),
        pytest.param(
            """\
            record(ai, "rec:X") {
                field(A, "test")
                field(B, "test")
            }
            alias("rec:X", "rec:Y")
            """,
            # standalone alisa -> record()
            """\
            record(ai, "rec:X") {
                alias("rec:Y")
                field(A, "test")
                field(B, "test")
            }
            """,
            id="alias_fix"
        ),
    ],
)
def test_roundtrip_format(input_text: str, expected: Optional[str]):
    input_text = textwrap.dedent(input_text)
    ctx = FormatContext(options=FormatOptions(indent=4))

    db = Database.from_string(input_text, version=3)
    result = ctx.render_object(db, "file")
    if expected is None:
        assert result.strip() == input_text.strip()
    else:
        expected = textwrap.dedent(expected)
        assert result.strip() == expected.strip()



"""
Generic parsing tools, for parsing when you know a file name or file type.
"""

import asyncio
import logging
import pathlib
import sys
from typing import Dict, Optional, Union

from .access_security import AccessSecurityConfig
from .common import AnyPath, FileFormat, IocMetadata
from .db import Database
from .dbtemplate import TemplateSubstitution
from .gateway import PVList as GatewayPVList
from .macro import MacroContext, PassthroughMacroContext
from .makefile import Makefile
from .shell import LoadedIoc
from .snl import SequencerProgram
from .streamdevice import StreamProtocol

logger = logging.getLogger(__name__)


ParseResult = Union[
    AccessSecurityConfig,
    Database,
    GatewayPVList,
    LoadedIoc,
    SequencerProgram,
    StreamProtocol,
    TemplateSubstitution,
    Makefile,
]


def parse(
    filename: AnyPath,
    dbd: Optional[str] = None,
    standin_directories: Optional[Dict[str, str]] = None,
    macros: Optional[str] = None,
    disable_macros: bool = False,
    use_gdb: bool = False,
    format: Optional[FileFormat] = None,
    expand: bool = False,
    v3: bool = False,
) -> ParseResult:
    """
    Generically parse either a startup script or a database file.

    Hopefully does the right thing based on file extension.  If not, use
    the ``format`` kwarg to specify it directly.

    Parameters
    ----------
    filename : str or pathlib.Path
        The filename to parse.

    dbd : str or pathlib.Path, optional
        The associated database definition file, if parsing a database or
        substitutions file.

    standin_directories : dict, optional
        Rewrite hard-coded directory prefixes by setting::

            standin_directories = {"/replace_this/": "/with/this"}

    macros : str, optional
        Macro string to use when parsing the file.

    disable_macros : bool, optional
        Disable macro handling, leaving unexpanded macros in the output.

    expand : bool, optional
        Expand a substitutions file.

    v3 : bool, optional
        Use V3 database grammar where applicable.
    """
    standin_directories = standin_directories or {}

    if str(filename) == "-":
        if format is None:
            options = [fmt.name for fmt in list(FileFormat)]
            raise ValueError(
                f"Format must be specified when piping data from standard "
                f"input.  Choose one of: {options}"
            )
        if format == FileFormat.iocsh:
            raise ValueError(
                "IOC shell script parsing not supported through standard input"
            )

        contents = sys.stdin.read()
        filename = pathlib.Path(sys.stdin.name)
    else:
        with open(filename, "rt") as fp:
            contents = fp.read()
        filename = pathlib.Path(filename)

    # The shared macro context - used in different ways below:
    if disable_macros:
        macro_context = PassthroughMacroContext()
    else:
        macro_context = MacroContext(macro_string=macros or "")

    if format is None:
        format = FileFormat.from_filename(filename)

    if format in (FileFormat.database, FileFormat.database_definition):
        if format == FileFormat.database_definition or not dbd:
            return Database.from_string(
                contents,
                filename=filename,
                macro_context=macro_context,
                version=3 if v3 else 4
            )
        return Database.from_string(
            contents,
            filename=filename,
            dbd=Database.from_file(dbd, version=3 if v3 else 4),
            macro_context=macro_context
        )

    if format == FileFormat.iocsh:
        md = IocMetadata.from_filename(
            filename,
            standin_directories=standin_directories,
            macros=dict(macro_context),
        )
        if use_gdb:
            try:
                asyncio.run(md.get_binary_information())
            except KeyboardInterrupt:
                logger.info("Skipping gdb information...")

        return LoadedIoc.from_metadata(md)

    if format == FileFormat.substitution:
        template = TemplateSubstitution.from_string(contents, filename=filename)
        if not expand:
            return template

        database_text = template.expand_files()
        # It's technically possible that this *isn't* a database file; so
        # perhaps a `whatrecord msi` could be implemented in the future.
        return Database.from_string(
            database_text,
            macro_context=macro_context,
            dbd=Database.from_file(dbd) if dbd is not None else None,
            filename=filename,
            version=3 if v3 else 4,
        )

    if format == FileFormat.state_notation:
        return SequencerProgram.from_string(contents, filename=filename)

    if macros:
        contents = macro_context.expand_file(contents)

    if format == FileFormat.gateway_pvlist:
        return GatewayPVList.from_string(contents, filename=filename)

    if format == FileFormat.access_security:
        return AccessSecurityConfig.from_string(contents, filename=filename)

    if format == FileFormat.stream_protocol:
        return StreamProtocol.from_string(contents, filename=filename)

    if format == FileFormat.makefile:
        return Makefile.from_string(contents, filename=filename)

    raise RuntimeError(
        f"Sorry, whatrecord doesn't support the {format!r} format just yet in the "
        f"CLI parsing tool.  Please open an issue."
    )



import pytest

from ..access_security import (AccessSecurityConfig, AccessSecurityGroup,
                               AccessSecurityRule, HostAccessGroup,
                               UserAccessGroup)
from ..common import LoadContext
from .conftest import MODULE_PATH

ACF_FILES = MODULE_PATH.glob("*.acf")

acf_files = pytest.mark.parametrize(
    "acf_file",
    [
        pytest.param(
            acf_file,
            id=acf_file.name,
        )
        for acf_file in ACF_FILES
    ]
)


def test_hag_simple():
    acf = AccessSecurityConfig.from_string("HAG(group) {host1, host2, host3}")
    assert acf.hosts["group"] == HostAccessGroup(
        context=(LoadContext("None", 1),),
        name="group",
        comments="",
        hosts=["host1", "host2", "host3"],
    )


def test_hag_simple_comments():
    acf = AccessSecurityConfig.from_string("""
# My group
# Information
HAG(group) {host1, host2, host3}
""")
    assert acf.hosts["group"] == HostAccessGroup(
        context=(LoadContext("None", 4),),
        name="group",
        comments="My group\nInformation",
        hosts=["host1", "host2", "host3"],
    )


def test_uag_simple():
    acf = AccessSecurityConfig.from_string(
        """
UAG(op) {op1,op2,superguy}
UAG(opSup) {superguy}
        """
    )

    assert acf.users["op"] == UserAccessGroup(
        context=(LoadContext("None", 2),),
        name="op",
        comments="",
        users=["op1", "op2", "superguy"],
    )

    assert acf.users["opSup"] == UserAccessGroup(
        context=(LoadContext("None", 3),),
        name="opSup",
        comments="",
        users=["superguy"],
    )


def test_uag_simple_comments():
    acf = AccessSecurityConfig.from_string(
        """
# Comments 1
UAG(op) {op1,op2,superguy}
# Comments 2
UAG(opSup) {superguy}
        """
    )

    assert acf.users["op"] == UserAccessGroup(
        context=(LoadContext("None", 3),),
        name="op",
        comments="Comments 1",
        users=["op1", "op2", "superguy"],
    )

    assert acf.users["opSup"] == UserAccessGroup(
        context=(LoadContext("None", 5),),
        name="opSup",
        comments="Comments 2",
        users=["superguy"],
    )


def test_asg_simple():
    acf = AccessSecurityConfig.from_string(
        """
    ASG(permit) {
        RULE(0,WRITE) {
            UAG(opSup,linacSup,appDev)
            HAG(host1, host2)
        }
        RULE(1,READ)
        RULE(1,WRITE) {
            HAG(ioc)
        }
    }
    """
    )

    assert acf.groups["permit"] == AccessSecurityGroup(
        context=(LoadContext("None", 2),),
        name="permit",
        comments="",
        rules=[
            AccessSecurityRule(
                context=(LoadContext("None", 3),),
                level=0,
                options="WRITE",
                comments="",
                log_options=None,
                users=["opSup", "linacSup", "appDev"],
                hosts=["host1", "host2"],
            ),
            AccessSecurityRule(
                context=(LoadContext("None", 7),),
                level=1,
                options="READ",
                comments="",
                log_options=None,
            ),
            AccessSecurityRule(
                context=(LoadContext("None", 8),),
                level=1,
                options="WRITE",
                comments="",
                log_options=None,
                hosts=["ioc"],
            ),
        ],
    )


def test_asg_log_options():
    acf = AccessSecurityConfig.from_string(
        """
# Comments
ASG(RWINSTRMCC) {
    # Comments 2
    RULE(1,WRITE,TRAPWRITE){
      HAG(drphosts,xpphosts)
    }
}
    """
    )

    assert acf.groups["RWINSTRMCC"] == AccessSecurityGroup(
        context=(LoadContext("None", 3),),
        name="RWINSTRMCC",
        comments="Comments",
        rules=[
            AccessSecurityRule(
                context=(LoadContext("None", 5),),
                level=1,
                options="WRITE",
                comments="Comments 2",
                log_options="TRAPWRITE",
                hosts=["drphosts", "xpphosts"],
            ),
        ],
    )


def test_asg_calc():
    acf = AccessSecurityConfig.from_string(
        """
ASG(DEFAULT) {
    INPA(LI:OPSTATE)
    INPB(LI:lev1permit)
    RULE(0,WRITE) {
        UAG(op)
        HAG(icr,cr)
        CALC("A=1")
    }
    RULE(0,WRITE) {
        UAG(op,linac,appdev)
        HAG(icr,cr)
        CALC("A=0")
    }
    RULE(1,WRITE) {
        UAG(opSup,linacSup,appdev)
        CALC("B=1")
    }
    RULE(1,READ)
    RULE(1,WRITE) {
        HAG(ioc)
    }
}
    """
    )

    assert acf.groups["DEFAULT"] == AccessSecurityGroup(
        context=(LoadContext("None", 2),),
        name="DEFAULT",
        comments="",
        inputs={"INPA": "LI:OPSTATE", "INPB": "LI:lev1permit"},
        rules=[
            AccessSecurityRule(
                context=(LoadContext("None", 5),),
                level=0,
                options="WRITE",
                log_options=None,
                users=["op"],
                hosts=["icr", "cr"],
                calc="A=1",
                comments="",
            ),
            AccessSecurityRule(
                context=(LoadContext("None", 10),),
                level=0,
                options="WRITE",
                log_options=None,
                users=["op", "linac", "appdev"],
                hosts=["icr", "cr"],
                calc="A=0",
                comments="",
            ),
            AccessSecurityRule(
                context=(LoadContext("None", 15),),
                level=1,
                options="WRITE",
                log_options=None,
                users=["opSup", "linacSup", "appdev"],
                calc="B=1",
                comments="",
            ),
            AccessSecurityRule(
                context=(LoadContext("None", 19),),
                level=1,
                options="READ",
                comments="",
            ),
            AccessSecurityRule(
                context=(LoadContext("None", 20),),
                level=1,
                options="WRITE",
                hosts=["ioc"],
                comments="",
            ),
        ],
    )


@acf_files
def test_load_full_acf_file_smoke(acf_file):
    loaded = AccessSecurityConfig.from_file(acf_file)
    # This is a smoke test - take a look at the output if interested
    print(loaded)
    assert loaded is not None



# Auto-generated from R7-2-1-20-gda3bfab4
# May not be backward/forward-compatible or 100% accurate

from dataclasses import dataclass
from typing import ClassVar, Dict

from . import asyn
from .common import AsynPortBase, ShellStateHandler

_handler = ShellStateHandler.generic_handler_decorator


@dataclass
class MotorState(ShellStateHandler):
    """Motor record support IOC shell state handler / container."""

    metadata_key: ClassVar[str] = "motor"

    @property
    def asyn(self) -> asyn.AsynState:
        """Asyn instance."""
        if self.primary_handler is None:
            raise RuntimeError("Requires a primary handler")
        return self.primary_handler.asyn

    @property
    def ports(self) -> Dict[str, AsynPortBase]:
        """Asyn ports."""
        return self.asyn.ports

    @_handler(stub=True)
    def handle_A3200AsynConfig(
        self,
        card_being_configured: int,
        asyn_port_name: str,
        asyn_address_gpib: int,
        number_of_axes: int,
        moving_poll_rate: int,
        idle_poll_rate: int,
        task_number: int,
        linear_move_commands: int,
    ):
        ...

    @_handler(stub=True)
    def handle_A3200AsynSetup(self, max_controller_count: int):
        ...

    @_handler(stub=True)
    def handle_ACRCreateController(
        self,
        port_name: str,
        acr_port_name: str,
        number_of_axes: int,
        moving_poll_period_ms: int,
        idle_poll_period_ms: int,
    ):
        ...

    @_handler(stub=True)
    def handle_AG_CONEXCreateController(
        self,
        port_name: str,
        serial_port_name: str,
        controller_id: int,
        moving_poll_period_ms: int,
        idle_poll_period_ms: int,
    ):
        ...

    @_handler(stub=True)
    def handle_AG_UCCreateAxis(
        self,
        controller_port_name: str,
        axis_number: int,
        has_limits: int,
        forward_amplitude: int,
        reverse_amplitude: int,
    ):
        ...

    @_handler(stub=True)
    def handle_AG_UCCreateController(
        self,
        port_name: str,
        serial_port_name: str,
        number_of_axes: int,
        moving_poll_period_ms: int,
        idle_poll_period_ms: int,
    ):
        ...

    @_handler(stub=True)
    def handle_ANC150AsynConfig(
        self,
        card_being_configured: int,
        asyn_port_name: str,
        number_of_axes: int,
        moving_poll_rate: int,
        idle_poll_rate: int,
    ):
        ...

    @_handler(stub=True)
    def handle_ANC150AsynSetup(self, maximum_of_controllers: int):
        ...

    @_handler(stub=True)
    def handle_ANF2CreateAxis(
        self,
        port_name: str,
        axis_number: int,
        hex_config: str,
        base_speed: int,
        homing_timeout: int,
    ):
        ...

    @_handler(stub=True)
    def handle_ANF2CreateController(
        self,
        port_name: str,
        anf2_in_port_name: str,
        anf2_out_port_name: str,
        number_of_axes: int,
    ):
        ...

    @_handler(stub=True)
    def handle_ANF2StartPoller(
        self, port_name: str, moving_poll_period_ms: int, idle_poll_period_ms: int
    ):
        ...

    @_handler(stub=True)
    def handle_ANG1CreateController(
        self,
        port_name: str,
        ang1_in_port_name: str,
        ang1_out_port_name: str,
        number_of_axes: int,
        moving_poll_period_ms: int,
        idle_poll_period_ms: int,
    ):
        ...

    @_handler(stub=True)
    def handle_AcsMotionConfig(
        self,
        acs_port_name: str,
        asyn_port_name: str,
        num_axes: int,
        moving_polling_rate: float,
        idle_polling_rate: float,
    ):
        ...

    @_handler(stub=True)
    def handle_C300CreateController(
        self,
        port_name: str,
        c300_port_name: str,
        number_of_axes: int,
        moving_poll_period_ms: int,
        idle_poll_period_ms: int,
    ):
        ...

    @_handler(stub=True)
    def handle_EMC18011Config(self, card_being_configured: int, asyn_port_name: str):
        ...

    @_handler(stub=True)
    def handle_EMC18011Setup(self, max_controller_count: int, polling_rate: int):
        ...

    @_handler(stub=True)
    def handle_ESP300Config(
        self, card_being_configured: int, asyn_port_name: str, asyn_address_gpib: int
    ):
        ...

    @_handler(stub=True)
    def handle_ESP300Setup(self, max_controller_count: int, polling_rate: int):
        ...

    @_handler(stub=True)
    def handle_EnsembleAsynConfig(
        self,
        card_being_configured: int,
        asyn_port_name: str,
        asyn_address_gpib: int,
        number_of_axes: int,
        moving_poll_rate: int,
        idle_poll_rate: int,
    ):
        ...

    @_handler(stub=True)
    def handle_EnsembleAsynSetup(self, max_controller_count: int):
        ...

    @_handler(stub=True)
    def handle_HXPCreateController(
        self,
        port_name: str,
        ip_address: str,
        port: int,
        moving_poll_period_ms: int,
        idle_poll_period_ms: int,
    ):
        ...

    @_handler(stub=True)
    def handle_Hytec8601Configure(
        self,
        port_name: str,
        num_axes: int,
        moving_poll_period: int,
        idle_poll_period: int,
        cardnum: int,
        carrier: int,
        ipslot: int,
        vector: int,
        useencoder: int,
        encoder_ratio0: float,
        encoder_ratio1: float,
        encoder_ratio2: float,
        encoder_ratio3: float,
    ):
        ...

    @_handler(stub=True)
    def handle_IM483PLConfig(self, card_being_configured: int, asyn_port_name: str):
        ...

    @_handler(stub=True)
    def handle_IM483PLSetup(self, max_controller_count: int, polling_rate: int):
        ...

    @_handler(stub=True)
    def handle_IM483SMConfig(self, card_being_configured: int, asyn_port_name: str):
        ...

    @_handler(stub=True)
    def handle_IM483SMSetup(self, max_controller_count: int, polling_rate: int):
        ...

    @_handler(stub=True)
    def handle_ImsMDrivePlusCreateController(
        self,
        motor_port_name: str,
        io_port_name: str,
        device_name: str,
        moving_poll_period_ms: float,
        idle_poll_period_ms: float,
    ):
        ...

    @_handler(stub=True)
    def handle_LinmotCreateController(
        self,
        port_name: str,
        linmot_port_name: str,
        number_of_axes: int,
        moving_poll_period_ms: int,
        idle_poll_period_ms: int,
    ):
        ...

    @_handler(stub=True)
    def handle_MAXvConfig(
        self,
        card_being_configured: int,
        configuration_string: str,
        absolute_encoder_flags: int,
        grey_code_flags: int,
    ):
        ...

    @_handler(stub=True)
    def handle_MAXvSetup(
        self,
        max_controller_count: int,
        vme_address_type: int,
        base_address_on_4_k_0x1000_boundary: int,
        valid_vectors: int,
        interrupt_level_1_6: int,
        polling_rate_hz: int,
    ):
        ...

    @_handler(stub=True)
    def handle_MCB4BConfig(self, card_being_configured: int, asyn_port_name: str):
        ...

    @_handler(stub=True)
    def handle_MCB4BCreateController(
        self,
        port_name: str,
        mcb_4_b_port_name: str,
        number_of_axes: int,
        moving_poll_period_ms: int,
        idle_poll_period_ms: int,
    ):
        ...

    @_handler(stub=True)
    def handle_MCB4BSetup(self, max_controller_count: int, polling_rate: int):
        ...

    @_handler(stub=True)
    def handle_MCDC2805Config(
        self,
        card_being_configured: int,
        modules_on_this_serial_port: int,
        asyn_port_name: str,
    ):
        ...

    @_handler(stub=True)
    def handle_MCDC2805Setup(self, max_controller_count: int, polling_rate: int):
        ...

    @_handler(stub=True)
    def handle_MCS2CreateController(
        self,
        port_name: str,
        mcs2_port_name: str,
        number_of_axes: int,
        moving_poll_period_ms: int,
        idle_poll_period_ms: int,
    ):
        ...

    @_handler(stub=True)
    def handle_MDT695Config(self, card_being_configured: int, asyn_port_name: str):
        ...

    @_handler(stub=True)
    def handle_MDT695Setup(self, max_controller_count: int, polling_rate: int):
        ...

    @_handler(stub=True)
    def handle_MDriveConfig(self, card_being_configured: int, asyn_port_name: str):
        ...

    @_handler(stub=True)
    def handle_MDriveSetup(self, max_controller_count: int, polling_rate: int):
        ...

    @_handler(stub=True)
    def handle_MM3000Config(
        self, card_being_configured: int, asyn_port_name: str, asyn_address_gpib: int
    ):
        ...

    @_handler(stub=True)
    def handle_MM300Setup(self, max_controller_count: int, polling_rate: int):
        ...

    @_handler(stub=True)
    def handle_MM4000AsynConfig(
        self,
        card_being_configured: int,
        asyn_port_name: str,
        asyn_address_gpib: int,
        number_of_axes: int,
        moving_poll_rate: int,
        idle_poll_rate: int,
    ):
        ...

    @_handler(stub=True)
    def handle_MM4000AsynSetup(self, max_controller_count: int):
        ...

    @_handler(stub=True)
    def handle_MM4000Config(
        self, card_being_configured: int, asyn_port_name: str, asyn_address_gpib: int
    ):
        ...

    @_handler(stub=True)
    def handle_MM4000Setup(self, max_controller_count: int, polling_rate: int):
        ...

    @_handler(stub=True)
    def handle_MMC200CreateController(
        self,
        port_name: str,
        mmc_200_port_name: str,
        number_of_axes: int,
        moving_poll_period_ms: int,
        idle_poll_period_ms: int,
        ignore_limit_flag: int,
    ):
        ...

    @_handler(stub=True)
    def handle_MVP2001CreateAxis(
        self,
        controller_port_name: str,
        axis_number: int,
        encoder_lines_per_rev: int,
        max_current_ma: int,
        limit_polarity: int,
    ):
        ...

    @_handler(stub=True)
    def handle_MVP2001CreateController(
        self,
        port_name: str,
        mvp_2001_port_name: str,
        number_of_axes: int,
        moving_poll_period_ms: int,
        idle_poll_period_ms: int,
    ):
        ...

    @_handler(stub=True)
    def handle_MXmotorSetup(self, max_motor: int, mx_data_file: str, polling_rate: int):
        ...

    @_handler(stub=True)
    def handle_MicosConfig(self, card_being_configured: int, asyn_port_name: str):
        ...

    @_handler(stub=True)
    def handle_MicosSetup(
        self, max_controller_count: int, max_motor_count: int, polling_rate: int
    ):
        ...

    @_handler(stub=True)
    def handle_MicroMoConfig(self, card_being_configured: int, asyn_port_name: str):
        ...

    @_handler(stub=True)
    def handle_MicroMoSetup(self, max_controller_count: int, polling_rate: int):
        ...

    @_handler(stub=True)
    def handle_OmsPC68Config(self, card_being_configured: int, asyn_port_name: str):
        ...

    @_handler(stub=True)
    def handle_OmsPC68Setup(self, maximum_of_cards: int, polling_rate_hz: int):
        ...

    @_handler(stub=True)
    def handle_PC6KConfig(self, card_being_configured: int, asyn_port_name: str):
        ...

    @_handler(stub=True)
    def handle_PC6KSetup(self, max_controller_count: int, polling_rate: int):
        ...

    @_handler(stub=True)
    def handle_PC6KUpLoad(
        self,
        controller_card: int,
        upload_file_path: str,
        program_name_null_immediate: str,
    ):
        ...

    @_handler(stub=True)
    def handle_PIC630Config(
        self,
        card_being_configured: int,
        asyn_port_name: str,
        ch_1_current_setting: int,
        ch_2_current_setting: int,
        ch_3_current_setting: int,
        ch_4_current_setting: int,
        ch_5_current_setting: int,
        ch_6_current_setting: int,
        ch_7_current_setting: int,
        ch_8_current_setting: int,
        ch_9_current_setting: int,
    ):
        ...

    @_handler(stub=True)
    def handle_PIC630Setup(
        self, max_controller_groups: int, max_axes_per_group: int, polling_rate: int
    ):
        ...

    @_handler(stub=True)
    def handle_PIC662Config(self, card_being_configured: int, asyn_port_name: str):
        ...

    @_handler(stub=True)
    def handle_PIC662Setup(self, max_controller_count: int, polling_rate: int):
        ...

    @_handler(stub=True)
    def handle_PIC663Config(
        self, card_being_configured: int, asyn_port_name: str, asyn_address_gpib: int
    ):
        ...

    @_handler(stub=True)
    def handle_PIC663Setup(self, max_controller_count: int, polling_rate: int):
        ...

    @_handler(stub=True)
    def handle_PIC844Config(
        self, card_being_configured: int, asyn_port_name: str, asyn_address_gpib: int
    ):
        ...

    @_handler(stub=True)
    def handle_PIC844Setup(self, max_controller_count: int, polling_rate: int):
        ...

    @_handler(stub=True)
    def handle_PIC848Config(
        self, card_being_configured: int, asyn_port_name: str, asyn_address_gpib: int
    ):
        ...

    @_handler(stub=True)
    def handle_PIC848Setup(self, max_controller_count: int, polling_rate: int):
        ...

    @_handler(stub=True)
    def handle_PIC862Config(
        self, card_being_configured: int, asyn_port_name: str, asyn_address_gpib: int
    ):
        ...

    @_handler(stub=True)
    def handle_PIC862Setup(self, max_controller_count: int, polling_rate: int):
        ...

    @_handler(stub=True)
    def handle_PIE516Config(
        self, card_being_configured: int, asyn_port_name: str, asyn_address_gpib: int
    ):
        ...

    @_handler(stub=True)
    def handle_PIE516Setup(self, max_controller_count: int, polling_rate: int):
        ...

    @_handler(stub=True)
    def handle_PIE517Config(
        self, card_being_configured: int, asyn_port_name: str, asyn_address_gpib: int
    ):
        ...

    @_handler(stub=True)
    def handle_PIE517Setup(self, max_controller_count: int, polling_rate: int):
        ...

    @_handler(stub=True)
    def handle_PIE710Config(
        self, card_being_configured: int, asyn_port_name: str, asyn_address_gpib: int
    ):
        ...

    @_handler(stub=True)
    def handle_PIE710Setup(self, max_controller_count: int, polling_rate: int):
        ...

    @_handler(stub=True)
    def handle_PIE816Config(
        self, card_being_configured: int, asyn_port_name: str, asyn_address_gpib: int
    ):
        ...

    @_handler(stub=True)
    def handle_PIE816Setup(self, max_controller_count: int, polling_rate: int):
        ...

    @_handler(stub=True)
    def handle_PIJEDSConfig(
        self, card_being_configured: int, asyn_port_name: str, asyn_address_gpib: int
    ):
        ...

    @_handler(stub=True)
    def handle_PIJEDSSetup(self, max_controller_count: int, polling_rate: int):
        ...

    @_handler(stub=True)
    def handle_PI_GCS2_CreateController(
        self,
        port_name: str,
        asyn_port_name: str,
        number_of_axes: int,
        priority: int,
        stack_size: int,
        moving_polling_time_msec: int,
        idle_polling_time_msec: int,
    ):
        ...

    @_handler(stub=True)
    def handle_PM304Config(
        self, card_being_configured: int, asyn_port_name: str, number_of_axes: int
    ):
        ...

    @_handler(stub=True)
    def handle_PM304Setup(self, max_controller_count: int, polling_rate: int):
        ...

    @_handler(stub=True)
    def handle_PM500Config(
        self, card_being_configured: int, asyn_port_name: str, asyn_address_gpib: int
    ):
        ...

    @_handler(stub=True)
    def handle_PM500Setup(self, max_controller_count: int, polling_rate: int):
        ...

    @_handler(stub=True)
    def handle_PMNC87xxConfig(
        self, card_being_configured: int, asyn_port_name: str, asyn_address_gpib: int
    ):
        ...

    @_handler(stub=True)
    def handle_PMNC87xxSetup(
        self,
        max_controller_count: int,
        max_drivers_per_controller_count: int,
        polling_rate: int,
    ):
        ...

    @_handler(stub=True)
    def handle_SC800Config(
        self, card_being_configured: int, asyn_port_name: str, asyn_address_gpib: int
    ):
        ...

    @_handler(stub=True)
    def handle_SC800Setup(self, maximum_of_cards: int, polling_rate_hz: int):
        ...

    @_handler(stub=True)
    def handle_SMC100Config(self, card_being_configured: int, asyn_port_name: str):
        ...

    @_handler(stub=True)
    def handle_SMC100CreateController(
        self,
        port_name: str,
        smc100_port_name: str,
        number_of_axes: int,
        moving_poll_period_ms: int,
        idle_poll_period_ms: int,
        eg_us_per_step: str,
    ):
        ...

    @_handler(stub=True)
    def handle_SMC100Setup(self, max_controller_count: int, polling_rate: int):
        ...

    @_handler(stub=True)
    def handle_SMCcorvusChangeResolution(
        self, smc_corvus_port_name: str, axis_number: int, axis_resolution: float
    ):
        ...

    @_handler(stub=True)
    def handle_SMCcorvusCreateController(
        self,
        port_name: str,
        smc_corvus_port_name: str,
        number_of_axes: int,
        moving_poll_period_ms: int,
        idle_poll_period_ms: int,
    ):
        ...

    @_handler(stub=True)
    def handle_SMChydraChangeResolution(
        self, smc_hydra_port_name: str, axis_number: int, axis_resolution: float
    ):
        ...

    @_handler(stub=True)
    def handle_SMChydraCreateController(
        self,
        port_name: str,
        smc_hydra_port_name: str,
        number_of_axes: int,
        moving_poll_period_ms: int,
        idle_poll_period_ms: int,
    ):
        ...

    @_handler(stub=True)
    def handle_SPiiPlusConfig(
        self,
        card_being_configured: int,
        asyn_port_name: str,
        command_mode_bu_ffer_co_nnect_di_rect: str,
    ):
        ...

    @_handler(stub=True)
    def handle_SPiiPlusSetup(self, max_controller_count: int, polling_rate: int):
        ...

    @_handler(stub=True)
    def handle_ScriptAxisConfig(
        self, controller_port_name: str, axis_number: int, parameters: str
    ):
        ...

    @_handler(stub=True)
    def handle_ScriptControllerConfig(
        self,
        motor_port_name: str,
        number_of_axes: int,
        control_script: str,
        parameters: str,
    ):
        ...

    @_handler(stub=True)
    def handle_ScriptMotorReload(self, motor_port_name: str):
        ...

    @_handler(stub=True)
    def handle_SmartMotorConfig(self, card_being_configured: int, asyn_port_name: str):
        ...

    @_handler(stub=True)
    def handle_SmartMotorSetup(self, max_controller_count: int, polling_rate: int):
        ...

    @_handler(stub=True)
    def handle_SoloistConfig(
        self, card_being_configured: int, asyn_port_name: str, asyn_address: int
    ):
        ...

    @_handler(stub=True)
    def handle_SoloistSetup(self, max_controller_count: int, polling_rate: int):
        ...

    @_handler(stub=True)
    def handle_XPSAuxConfig(
        self, port_name: str, ip_address: str, ip_port: int, polling_period: int
    ):
        ...

    @_handler(stub=True)
    def handle_XPSConfig(
        self,
        card_being_configured: int,
        ip: str,
        port: int,
        number_of_axes: int,
        moving_poll_rate: int,
        idle_poll_rate: int,
    ):
        ...

    @_handler(stub=True)
    def handle_XPSConfigAxis(
        self,
        card_number: int,
        axis_number: int,
        axis_name: str,
        steps_per_unit: str,
        no_disabled_error: int,
    ):
        ...

    @_handler(stub=True)
    def handle_XPSCreateAxis(
        self,
        controller_port_name: str,
        axis_number: int,
        axis_name: str,
        steps_per_unit: str,
    ):
        ...

    @_handler(stub=True)
    def handle_XPSCreateController(
        self,
        controller_port_name: str,
        ip_address: str,
        ip_port: int,
        number_of_axes: int,
        moving_poll_rate_ms: int,
        idle_poll_rate_ms: int,
        enable_set_position: int,
        set_position_settling_time_ms: int,
    ):
        ...

    @_handler(stub=True)
    def handle_XPSCreateProfile(
        self,
        controller_port_name: str,
        max_points: int,
        ftp_username: str,
        ftp_password: str,
    ):
        ...

    @_handler(stub=True)
    def handle_XPSDisableAutoEnable(self, controller_port_name: str):
        ...

    @_handler(stub=True)
    def handle_XPSDisablePoll(self, set_disable_poll_value: int):
        ...

    @_handler(stub=True)
    def handle_XPSEnableMoveToHome(
        self, card_number: int, axis_name: str, distance: int
    ):
        ...

    @_handler(stub=True)
    def handle_XPSEnableMovingMode(self, controller_port_name: str):
        ...

    @_handler(stub=True)
    def handle_XPSEnableSetPosition(self, set_position_flag: int):
        ...

    @_handler(stub=True)
    def handle_XPSGathering(self, interelement_period: int):
        ...

    @_handler(stub=True)
    def handle_XPSInterpose(self, port_name: str):
        ...

    @_handler(stub=True)
    def handle_XPSNoDisableError(self, controller_port_name: str):
        ...

    @_handler(stub=True)
    def handle_XPSSetPosSleepTime(self, set_position_sleep_time: int):
        ...

    @_handler(stub=True)
    def handle_XPSSetup(self, number_of_xps_controllers: int):
        ...

    @_handler(stub=True)
    def handle_asynMotorEnableMoveToHome(
        self, controller_port_name: str, axis_number: int, distance: int
    ):
        ...

    @_handler(stub=True)
    def handle_listMovingMotors(self, list_moving_motors: str):
        ...

    @_handler(stub=True)
    def handle_motorSimConfigAxis(
        self,
        post_name: str,
        axis: int,
        high_limit: int,
        low_limit: int,
        home_position: int,
        start_posn: int,
    ):
        ...

    @_handler(stub=True)
    def handle_motorSimCreate(
        self,
        card: int,
        signal: int,
        high_limit: int,
        low_limit: int,
        home_position: int,
        num_cards: int,
        num_signals: int,
        start_posn: int,
    ):
        ...

    @_handler(stub=True)
    def handle_motorSimCreateController(
        self, port_name: str, number_of_axes: int, priority: int, stack_size: int
    ):
        ...

    @_handler(stub=True)
    def handle_motorUtilInit(self, ioc_name: str):
        ...

    @_handler(stub=True)
    def handle_nf874xCreateController(
        self,
        port_name: str,
        nf874x_port_name: str,
        number_of_axes: int,
        moving_poll_period_ms: int,
        idle_poll_period_ms: int,
    ):
        ...

    @_handler(stub=True)
    def handle_oms58Setup(
        self, num_card: int, addrs: int, vector: int, int_level: int, scan_rate: int
    ):
        ...

    @_handler(stub=True)
    def handle_omsMAXnetConfig(
        self,
        asyn_motor_port_name: str,
        number_of_axes: int,
        asyn_serial_tcp_port_name: str,
        moving_poll_rate: int,
        idle_poll_rate: int,
        initstring: str,
    ):
        ...

    @_handler(stub=True)
    def handle_omsMAXvConfig(
        self,
        number_of_card: int,
        asyn_motor_port_name: str,
        number_of_axes: int,
        moving_poll_rate: int,
        idle_poll_rate: int,
        initstring: str,
    ):
        ...

    @_handler(stub=True)
    def handle_omsMAXvConfig2(
        self,
        slot_number: int,
        address_type_a16_a24_a32: str,
        board_address_on_4_k_0x1000_boundary: int,
        interrupt_vector_noninterrupting_0_64_255: int,
        interrupt_level_1_6: int,
        asyn_motor_port_name: str,
        number_of_axes: int,
        task_priority_0_medium: int,
        stack_size_0_medium: int,
        moving_poll_rate: int,
        idle_poll_rate: int,
        initstring: str,
    ):
        ...

    @_handler(stub=True)
    def handle_omsMAXvEncFuncConfig(
        self,
        number_of_card: int,
        asyn_motor_port_name: str,
        number_of_axes: int,
        moving_poll_rate: int,
        idle_poll_rate: int,
        initstring: str,
    ):
        ...

    @_handler(stub=True)
    def handle_omsMAXvEncFuncConfig2(
        self,
        slot_number: int,
        address_type_a16_a24_a32: str,
        board_address_on_4_k_0x1000_boundary: int,
        interrupt_vector_noninterrupting_0_64_255: int,
        interrupt_level_1_6: int,
        asyn_motor_port_name: str,
        number_of_axes: int,
        task_priority_0_medium: int,
        stack_size_0_medium: int,
        moving_poll_rate: int,
        idle_poll_rate: int,
        initstring: str,
    ):
        ...

    @_handler(stub=True)
    def handle_omsMAXvSetup(
        self,
        max_controller_count: int,
        vme_address_type: int,
        base_address_on_4_k_0x1000_boundary: int,
        noninterrupting_0_valid_vectors_64_255: int,
        interrupt_level_1_6: int,
    ):
        ...

    @_handler(stub=True)
    def handle_phytronCreateAxis(
        self, controller_name: str, module_index: int, axis_index: int
    ):
        ...

    @_handler(stub=True)
    def handle_phytronCreateController(
        self,
        port_name: str,
        phytron_axis_port_name: str,
        moving_poll_period_ms: int,
        idle_poll_period_ms: int,
        timeout_ms: float,
    ):
        ...

    @_handler(stub=True)
    def handle_printChIDlist(self, print_motor_util_chid_list: str):
        ...

    @_handler(stub=True)
    def handle_setIdlePollPeriod(self, controller_port_name: str, axis_number: float):
        ...

    @_handler(stub=True)
    def handle_setMovingPollPeriod(self, controller_port_name: str, axis_number: float):
        ...

    @_handler(stub=True)
    def handle_smarActMCSCreateAxis(
        self, controller_port_name_string: str, axis_number_int: int, channel_int: int
    ):
        ...

    @_handler(stub=True)
    def handle_smarActMCSCreateController(
        self,
        port_name_string: str,
        i_o_port_name_string: str,
        number_of_axes_int: int,
        moving_poll_period_s_double: float,
        idle_poll_period_s_double: float,
    ):
        ...

    @_handler(stub=True)
    def handle_smarActSCUCreateAxis(
        self, controller_port_name_string: str, axis_number_int: int, channel_int: int
    ):
        ...

    @_handler(stub=True)
    def handle_smarActSCUCreateController(
        self,
        port_name_string: str,
        i_o_port_name_string: str,
        number_of_axes_int: int,
        moving_poll_period_s_double: float,
        idle_poll_period_s_double: float,
    ):
        ...

    @_handler(stub=True)
    def handle_tclcall(self, tcl_name: str, task_name: str, function_args: str):
        ...

    @_handler(stub=True)
    def handle_xps_gathering(self, element_period_10_4: int):
        ...

    @_handler(stub=True)
    def handle_EthercatMCCreateAxis(
        self, motor_port: str, axis_num: int, amplifier_flags: str, axis_config: str
    ):
        ...

    @_handler
    def handle_adsAsynPortDriverConfigure(
        self,
        portName: str,
        ipaddr: str = "",
        amsaddr: str = "",
        amsport: int = 0,
        asynParamTableSize: int = 0,
        priority: int = 0,
        noAutoConnect: int = 0,
        defaultSampleTimeMS: int = 0,
        maxDelayTimeMS: int = 0,
        adsTimeoutMS: int = 0,
        defaultTimeSource: str = "",
    ):
        # SLAC-specific, but doesn't hurt anyone
        self.ports[portName] = asyn.AdsAsynPort(
            context=self.get_load_context(),
            name=portName,
            ipaddr=ipaddr,
            amsaddr=amsaddr,
            amsport=amsport,
            asynParamTableSize=asynParamTableSize,
            priority=priority,
            noAutoConnect=noAutoConnect,
            defaultSampleTimeMS=defaultSampleTimeMS,
            maxDelayTimeMS=maxDelayTimeMS,
            adsTimeoutMS=adsTimeoutMS,
            defaultTimeSource=defaultTimeSource,
        )

    @_handler
    def handle_drvAsynMotorConfigure(
        self,
        port_name: str = "",
        driver_name: str = "",
        card_num: int = 0,
        num_axes: int = 0,
    ):
        self.ports[port_name] = asyn.AsynMotor(
            context=self.get_load_context(),
            name=port_name,
            parent=None,
            metadata=dict(
                num_axes=num_axes,
                card_num=card_num,
                driver_name=driver_name,
            ),
        )

    @_handler
    def handle_EthercatMCCreateController(
        self,
        motor_port: str,
        asyn_port: str,
        num_axes: int = 0,
        move_poll_rate: float = 0.0,
        idle_poll_rate: float = 0.0,
    ):
        # SLAC-specific
        port = self.ports[asyn_port]
        motor = asyn.AsynMotor(
            context=self.get_load_context(),
            name=motor_port,
            parent=asyn_port,
            metadata=dict(
                num_axes=num_axes,
                move_poll_rate=move_poll_rate,
                idle_poll_rate=idle_poll_rate,
            ),
        )

        # Tie it to both the original asyn port (as a motor) and also the
        # top-level asyn ports.
        port.motors[motor_port] = motor
        self.ports[motor_port] = motor



import os
import pathlib

import pytest

from .. import makefile, settings

MODULE_PATH = pathlib.Path(__file__).resolve().parent

STARTUP_SCRIPTS = list((MODULE_PATH / "iocs").glob("**/st.cmd"))

# Disable caching for _all_ tests
settings.CACHE_PATH = ""
os.environ["WHATRECORD_CACHE_PATH"] = ""

startup_scripts = pytest.mark.parametrize(
    "startup_script",
    [
        pytest.param(
            startup_script,
            id="/".join(startup_script.parts[-2:])
        )
        for startup_script in STARTUP_SCRIPTS
    ]
)

skip_without_make = pytest.mark.skipif(
    not makefile.host_has_make(),
    reason="Host does not have make"
)



"""
"whatrecord iocmanager-loader" is used to import IOC configuration information
from SLAC PCDS IocManager-format configuration files.
"""

import argparse
import json
import logging
from typing import List

from ..common import AnyPath, IocInfoDict
from ..iocmanager import get_iocs_from_configs

logger = logging.getLogger(__name__)
DESCRIPTION = __doc__


def build_arg_parser(parser=None):
    if parser is None:
        parser = argparse.ArgumentParser()

    parser.description = DESCRIPTION
    parser.formatter_class = argparse.RawTextHelpFormatter

    parser.add_argument(
        "configs", type=str, nargs="+", help="Configuration file location(s)"
    )

    parser.add_argument(
        "--limit", type=int, default=0, help="Limit the number of IOCs to output"
    )

    return parser


def main(configs: list[AnyPath], limit: int = 0) -> List[IocInfoDict]:
    iocs = get_iocs_from_configs(configs)
    if limit:
        dumped = json.dumps(iocs[-limit:], indent=4)
    else:
        dumped = json.dumps(iocs, indent=4)

    print(dumped)
    return iocs



from typing import Dict, List, Optional, Union

import aiohttp
import apischema

from . import settings
from .server.server import IocGetMatchesResponse, PVGetInfo

Params = Dict[str, Union[List[str], str]]


async def make_query(path, server=None, params: Optional[Params] = None):
    params = params or {}
    server = (server or settings.WHATREC_SERVER).rstrip("/")

    if not path.startswith("/"):
        raise ValueError(
            "Expects path not including server, and should start with ``/``"
        )

    async with aiohttp.ClientSession() as session:
        async with session.get(f"{server}{path}", params=params) as resp:
            if resp.status != 200:
                raise RuntimeError(f"Server error: {resp.status}")
            return await resp.json()


async def get_record_info(*records, server: Optional[str] = None) -> Dict[str, PVGetInfo]:
    """Get record information from the server."""
    response = await make_query("/api/pv/info", server=server, params=dict(pv=list(records)))
    return apischema.deserialize(Dict[str, PVGetInfo], response)


async def get_iocs(
    pattern: str = "*", server: Optional[str] = None, regex: bool = False
) -> IocGetMatchesResponse:
    """Get record information from the server."""
    response = await make_query(
        "/api/ioc/matches", server=server, params=dict(pattern=pattern, regex=str(regex))
    )
    return apischema.deserialize(IocGetMatchesResponse, response)



import asyncio
import logging
import pathlib
import tempfile

import pytest

from ..server.server import ServerState, _new_server_state
from .conftest import MODULE_PATH, STARTUP_SCRIPTS

logger = logging.getLogger(__name__)


def test_init_basic():
    ServerState()


@pytest.fixture()
def state():
    return _new_server_state(
        scripts=[str(script) for script in STARTUP_SCRIPTS],
        gateway_config=str(MODULE_PATH),
    )


@pytest.fixture()
def ready_state(state: ServerState, caplog):
    caplog.set_level("INFO")
    # asyncio.run(state.async_init(app=None), debug=True)
    logger.info("Updating script loaders")
    asyncio.run(state.update_script_loaders())
    asyncio.run(state.update_iocs(state.get_updated_iocs()))
    state._load_gateway_config()
    # asyncio.run(state.update_plugins())
    return state


def test_cli_state_init(state: ServerState):
    ...


def test_misc(ready_state: ServerState):
    print(ready_state.ioc_metadata)
    assert ready_state._update_count == 1

    pvname, iocname = "IOC:KFE:A:One", "ioc_a"
    what, = ready_state.whatrec(pvname)
    assert what.record is not None
    assert what.pva_group is None
    assert what.record.instance == ready_state.database[pvname]

    script = ready_state.container.scripts[iocname]
    dbd = script.shell_state.database_definition
    assert what.record.definition == dbd.record_types[
        what.record.instance.record_type
    ]

    pvname, iocname = "PVACCESS:SIMPLE:01", "pva_simple"
    what, = ready_state.whatrec(pvname)
    assert what.record is None
    assert what.pva_group is not None
    assert what.pva_group == ready_state.pva_database[pvname]


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "format, full",
    [
        pytest.param(".json", False, id="uncompressed-partial"),
        pytest.param(".json.gz", True, id="compressed-full"),
    ]
)
async def test_dump_to_file(ready_state: ServerState, format: str, full: bool):
    with tempfile.NamedTemporaryFile(suffix=format) as fp:
        await ready_state.dump_to_file(filename=pathlib.Path(fp.name), full=full)
